'''
Created on 2020-09-15

@author: wf
'''
from corpus.eventcorpus import EventDataSource, EventDataSourceConfig
from corpus.event import EventStorage,EventSeriesManager, EventSeries, Event, EventManager
from lodstorage.storageconfig import StorageConfig
from datetime import datetime
import json
import re
from collections import Counter
from corpus.utils.textparse import Textparse

class GND(EventDataSource):
    '''
    manages event data from Gemeinsame Normdatei
    https://d-nb.info/standards/elementset/gnd
    '''
    debug=False
    host="confident.dbis.rwth-aachen.de"
    endpoint=f"https://{host}/jena/gnd/sparql"
    sourceConfig = EventDataSourceConfig(lookupId="gnd", name="GND", url='https://d-nb.info/standards/elementset/gnd', title='Gemeinsame Normdatei', tableSuffix="gnd",locationAttribute="location")
    
    def __init__(self,debug=False):
        '''
        constructor
        '''
        super().__init__(GndEventManager(), GndEventSeriesManager(), GND.sourceConfig)
        self.debug=debug
    
class ExtractStatistics:
    '''
    statistics 
    '''
    def __init__(self):
        self.counter=Counter()
        self.invalidTitles=[]
        self.invalidOrdinals=[]
        self.invalidYears=[]
        self.invalidDates=[]
        
    def dump(self):
        '''
        dump me
        '''
        timestamp=datetime.utcnow().strftime('%Y-%m-%d_%H:%M:%S')
        jsonfile = open(f"/tmp/extractstats_{timestamp}.json", "w")
        json.dump(self.__dict__, jsonfile, indent = 2)
        
    def addPartsLen(self,plen:int):
        self.counter[plen]+=1
        
    def addInvalidTitle(self,title):
        self.invalidTitles.append(title)
        self.counter["invalidTitle"]+=1
        
    def addInvalidOrdinal(self,ordinalStr):
        self.invalidOrdinals.append(ordinalStr)
        self.counter["invalidOrdinal"]+=1
       
    def addInvalidYear(self,yearStr): 
        self.invalidYears.append(yearStr)
        self.counter["invalidYear"]+=1
        
    def addInvalidDate(self,dateStr):
        self.invalidDates.append(dateStr)
        self.counter["invalidDate"]+=1
       
    
class GndTitleExtractor:
    
    def __init__(self,fulltitle:str,stats):
        '''
        constructor
        '''
        self.fulltitle=fulltitle
        self.stats=stats
        self._fields=["title","year","location","ordinal","organization"]
        
    def __str__(self):
        '''
        get my string representation
        '''
        text=""
        delim=""
        for field in self._fields:
            if hasattr(self, field):
                text+=f"{delim}{field}:{getattr(self,field)}" 
                delim="\n"
        return text
    
    def setOrganization(self,orgStr):
        self.organization=orgStr
    
    def setLocation(self,location):
        self.location=location.strip()
        
    def setOrdinal(self,ordinalStr):
        ordinalStr=ordinalStr.strip()
        m=re.match(r"^([0-9]+)\.?$",ordinalStr)
        if m:
            ordinal=int(m.group(1))
            self.ordinal=ordinal
            pass
        else:
            self.stats.addInvalidOrdinal(ordinalStr)
    
    def setYear(self,yearStr):
        '''
        set the year of the given event 
        
        Args:
            yearStr(str): the year string to analyze
        '''
        yearStr=yearStr.strip()
        if re.match(r"^[0-9]+$",yearStr):
            self.year=int(yearStr)
            return True
        else: 
            m=re.match(r"^([0-9]+)\s*\-\s*([0-9]+)$",yearStr)
            if m:
                self.year=int(m.group(1))
                self.stats.counter["yearRange"]+=1
                return True
            else:
                self.stats.addInvalidYear(yearStr)
                return False
     
    def titleExtract(self):
        '''
        extract meta data information from the given fulltitle of this gnd event
        
        Args:
            rawEvent(dict): the dict as queried
        '''
        regex=r"(.*?)\((.*?)\)"
        m=re.match(regex,self.fulltitle)
        stats=self.stats
        if m:
            self.title=m.group(1).strip()
            ordYearLocation=m.group(2)
            parts=ordYearLocation.split(":")
            plen=len(parts)
            stats.addPartsLen(plen)
            if plen==1:
                self.setYear(ordYearLocation)
            elif plen==2:
                self.setYear(parts[0])
                self.setLocation(parts[1])
            elif plen==3:
                self.setOrdinal(parts[0])
                self.setYear(parts[1])
                self.setLocation(parts[2])
            elif plen==4:
                self.setOrganization(parts[0])
                self.setOrdinal(parts[1])
                self.setYear(parts[2])
                self.setLocation(parts[3])
                pass
        else:
            stats.addInvalidTitle(self.fulltitle)
        
    def updateRawEvent(self,rawEvent:dict):
        '''
        update the given rawEvent with my values
        '''
        for field in self._fields:
            if hasattr(self, field):
                rawEvent[field]=getattr(self,field)
        
        
class GndEvent(Event):
    '''
    a scientific event derived from Gemeinsame Normdatei
    '''
    
    @classmethod
    def getSamples(cls):
        samples = [      
        ]
        return samples
    
            
    @classmethod
    def postProcessLodRecord(cls,rawEvent:dict):
        '''
        fix the given raw Event
        
        Args:
            rawEvent(dict): the raw event record to fix
        '''
      
class GndEventSeries(EventSeries):
    '''
    a scientific event series derived from Gemeinsame Normdatei
    '''
    
        
class GndEventManager(EventManager):
    '''
    manage WikiCFP derived scientific events
    '''

    def __init__(self, config:StorageConfig=None):
        '''
        Constructor
        '''
        self.source="gnd"
        self.endpoint=GND.endpoint
        self.queryManager=EventStorage.getQueryManager(lang="sparql",name="gnd")
        super().__init__(name="GndEvents", sourceConfig=GND.sourceConfig,clazz=GndEvent, config=config)
    
    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts=self.getLoDfromEndpoint
        
    def postProcessEntityList(self,debug:bool=False):
        '''
        postProcess Events
        '''
        stats=ExtractStatistics()
        for event in self.events:
            titleExtractor=GndTitleExtractor(getattr(event, "fulltitle", ""),stats)
            titleExtractor.titleExtract()
            titleExtractor.updateRawEvent(event.__dict__)
            dateRange=(Textparse.getDateRange(event.date))
            if (len(dateRange)==0 and event.date is not None):
                stats.addInvalidDate(event.date)
            else:
                Textparse.setDateRange(event, dateRange)
        print(stats.counter.most_common())
        if debug:
            stats.dump()
        return stats
    
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this event manager
        '''
        gndquery=self.queryManager.queriesByName["GND-Events"]
        sparql=gndquery.query
        return sparql


class GndEventSeriesManager(EventSeriesManager):
    '''
    manage GND derived scientific conference series
    '''

    def __init__(self, config:StorageConfig=None):
        '''
        Constructor
        '''
        self.source="gnd"
        self.endpoint=GND.endpoint
        super().__init__(name="GndEventSeries", sourceConfig=GND.sourceConfig, clazz=GndEventSeries, config=config)

    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            #self.getListOfDicts=self.getLoDfromEndpoint  
            self.getListOfDicts=self.getDummyLoD
            
    def getDummyLoD(self):
        lod=[
            {
                "seriesId":"0",
                "title":""
            }
        ]
        return lod
        
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this event series manager
        '''
        sparql="""SELECT ?s ?p ?o 
WHERE {
  VALUES (?s) { ("subject") }
  VALUES (?p) { ("predicate") }
  VALUES (?o) { ("object") }
}
        """
        return sparql
        


'''
Created on 2021-05-03

Proceedings Title signature elements
@author: wf
'''
from functools import reduce
from typing import Union

from ptp.parsing import Category
from num2words import num2words
import re
import yaml
import pathlib
from pyparsing import ParseException
import pyparsing as pp


class RegexpCategory(Category):
    '''
    category defined by a regular expression
    '''
    def __init__(self,name,itemFunc,pattern):
        '''
        constructor
        
        Args:
            name(str): the name of this category
            itemFunc(func): the function to use for getting values
            pattern(str): the regular expression pattern to apply
        '''
        self.pattern=pattern
        super().__init__(name,itemFunc=itemFunc)
        
    def checkMatch(self,token):
        return re.search(self.pattern, token) is not None
    
class ParsingCategory(Category):
    '''
    category defined by a PyParsing grammar
    '''
    def __init__(self,name,grammar):
        self.grammar=grammar
        itemFunc=self.getResult
        super().__init__(name,itemFunc=itemFunc)
    
    def getResult(self):
        pass
        return None
    
    def checkMatch(self,token):
        self.parseResult=None
        try:
            self.parseResult=self.grammar.parseString(token)
            pass
        except ParseException as _pe:
            # ignore any parse exception
            pass
        return self.parseResult is not None
        
  
class EnumCategory(Category):
    '''
    category defined by an enumeration specified in a lookup table
    '''
   
    def getYamlPath(self,lookupName:str):
        '''
        get the YamlPath for the given filename prefix
        
        Args:
            filename(str): the prefix of the filename to get the Yaml path for
            
        Returns:
            str: the string representation of the path
        '''
        grandParent = pathlib.Path(__file__).parent.parent.resolve()
        filepath = (grandParent / f"resources/{lookupName}.yaml")
        path=str(filepath)
        return path
  
    def read(self, yamlPath=None):
        ''' read the dictionary from the given yaml path
            https://github.com/WolfgangFahl/ProceedingsTitleParser/blob/7e52b4e3eae09269464669fe387425b9f6392952/ptp/titleparser.py#L456
        '''
        if yamlPath is None:
            yamlPath=self.getYamlPath()
        with open(yamlPath, 'r') as stream:
            self.tokens = yaml.safe_load(stream)
        pass
    
    def __init__(self,name:str, lookupName:str=None):
        '''
        construct me for the given name

        Args:
            name(str): name of the Category (type field in the yaml)
            tokenPath(str): path to the yaml file containing the tokens/enum information
        '''
        super().__init__(name,itemFunc=lambda word:self.lookup(word))
        self.lookupByKey={}   
        if lookupName is None:
            multiType=True
            lookupName="dictionary"
        else:
            multiType=False
        lookupYamlPath=self.getYamlPath(lookupName) 
        self.read(lookupYamlPath)
        # loop over all tokens
        for tokenKey in self.tokens:
            token=self.tokens[tokenKey]
            # get the token
            if multiType:
                tokenType=token['type']
                valid=tokenType==name
            else:
                valid=True
            if valid:
                value=token['value'] if 'value' in token else tokenKey 
                self.lookupByKey[tokenKey]=value
    
    def checkMatch(self,word):
        return self.lookup(word) is not None
    
    def lookup(self,word):
        '''
        lookup the given word
        '''
        if word in self.lookupByKey:
            return self.lookupByKey[word]
        return None
    
    def addLookup(self,key,value):
        '''
        add the given key value pair to my lookup
        '''
        self.lookupByKey[key]=value

class OrdinalCategory(EnumCategory):
    '''
    I am the category for ordinals
    '''

    def __init__(self,maxOrdinal=250):
        '''
        Constructor
        '''
        self.maxOrdinal=maxOrdinal
        super().__init__("Ordinal")
        self.prepareLookup()
        
    def prepareLookup(self):
        # https://stackoverflow.com/a/20007730/1497139
        ordinal = lambda n: "%d%s" % (n,"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4])
        for i in range(1,self.maxOrdinal):
            # standard decimal ordinal 1., 2., 3. 
            self.addLookup(f"{i}.",i)
            # text ordinal 1st, 2nd, 3rd
            ordinalStr=f"{ordinal(i)}"
            self.addLookup(ordinalStr,i)
            # roman
            self.addLookup(f"{self.toRoman(i)}.",i)
            # using num2words library
            ordinal4i=num2words(i, to='ordinal')
            self.addLookup(ordinal4i,i)
            title=ordinal4i.title()
            self.addLookup(title,i)
            pass       

    @classmethod
    def toRoman(cls,number):
        ''' https://stackoverflow.com/a/11749642/1497139 '''
        if (number < 0) or (number > 3999):
            raise Exception("number needs to be between 1 and 3999")
        if (number < 1): return ""
        if (number >= 1000): return "M" + cls.toRoman(number - 1000)
        if (number >= 900): return "CM" + cls.toRoman(number - 900)
        if (number >= 500): return "D" + cls.toRoman(number - 500)
        if (number >= 400): return "CD" + cls.toRoman(number - 400)
        if (number >= 100): return "C" + cls.toRoman(number - 100)
        if (number >= 90): return "XC" + cls.toRoman(number - 90)
        if (number >= 50): return "L" + cls.toRoman(number - 50)
        if (number >= 40): return "XL" + cls.toRoman(number - 40)
        if (number >= 10): return "X" + cls.toRoman(number - 10)
        if (number >= 9): return "IX" + cls.toRoman(number - 9)
        if (number >= 5): return "V" + cls.toRoman(number - 5)
        if (number >= 4): return "IV" + cls.toRoman(number - 4)
        if (number >= 1): return "I" + cls.toRoman(number - 1)


class CountryCategory(EnumCategory):
    '''
    I am the category for countries
    '''

    def __init__(self):
        '''
        constructor
        '''
        super().__init__("country", lookupName="countries")
        
class YearCategory(EnumCategory):
    '''
    I am the category for years
    '''

    def __init__(self):
        '''
        constructor
        '''
        super().__init__("year", lookupName="years")


class RomanNumeral(pp.OneOrMore):
    """
    matches roman numerals
    see https://github.com/pyparsing/pyparsing/blob/master/examples/romanNumerals.py
    limited to number up to 99
    """

    def __init__(self, limit:int=None, name:str=None):
        """
        constructor
        """
        if name is None:
            name = "roman"
        map = [("I", 1), ("IV", 4), ("V", 5), ("IX", 9), ("X", 10), ("XL", 40), ("L", 50), ("XC", 90), ("C", 100),
               ("CD", 400), ("D", 500), ("CM", 900), ("M", 1000)]
        expressions = [self.romanNumeralLiteral(numeralString, value) for numeralString, value in map if limit is None or value < limit]
        numeral = reduce(lambda a, b: b|a, expressions).leaveWhitespace()
        super(RomanNumeral, self).__init__(numeral[1, ...])
        self.setParseAction(sum)
        self.setName(name)

    @classmethod
    def romanNumeralLiteral(cls, numeralString, value):
        return pp.Literal(numeralString).setParseAction(pp.replaceWith(value))


class VolumeCategory(Category):
    """
    I am a category for volumes
    """
    VOLUME_LABEL = pp.Group(pp.CaselessLiteral("volume") |
                            pp.CaselessLiteral("vol.") |
                            pp.CaselessLiteral("vol") |
                            pp.CaselessLiteral("part") |
                            pp.CaselessLiteral("pt.")
                            )("label")
    VOLUME_VALUE_NUMERIC = pp.Word(pp.nums).setParseAction(pp.pyparsing_common.convertToInteger).setName("numeric")
    VOLUME_VALUE_ALPHA = pp.Char(pp.alphas).setParseAction(pp.tokenMap(lambda value: ord(value.lower())-96)).setName("alphanumeric")
    VOLUME_VALUE = pp.Group(VOLUME_VALUE_NUMERIC | RomanNumeral(limit=100, name="roman reduced") | VOLUME_VALUE_ALPHA)("value")
    VOLUME = VOLUME_LABEL + pp.Optional(" ") + VOLUME_VALUE

    def __init__(self):
        """
        constructor
        """
        super().__init__("volume", self.getValue)

    def checkMatch(self, word:str) -> bool:
        return self.VOLUME.matches(word)

    def checkMatchWithContext(self, tokenStr: str, pos: int, tokenSequence) -> bool:
        matches = False
        if self.VOLUME_LABEL.matches(tokenStr):
            matches = True
        else:
            previousTokens = tokenSequence.getTokenOfCategory(self.name)
            tokensAtPreviousPosition = [token for token in previousTokens if token.pos == pos-1]
            if len(tokensAtPreviousPosition) == 1:
                tokenAtPreviousPosition = tokensAtPreviousPosition[0]
                if tokenAtPreviousPosition.value == "LABEL":
                    # previous token is volume label → match against volume value
                    matches = self.VOLUME_VALUE.matches(tokenStr)
        return matches


    def getValue(self, word:str) -> Union[str, int, None]:
        """
        returns the value of the given word.
        LABEL if the word is the volume label
        <integer> if the word is the volume label
        Args:
            word: word to parse the value for

        Returns:
            value of the given word
        """
        res = None
        if self.VOLUME.matches(word):
            parsedValue = self.VOLUME.parseString(word).asDict()
            vals = [m for m,_,_ in self.VOLUME.scanString(word)]
            res = parsedValue.get("value")[0]
        elif self.VOLUME_LABEL.matches(word):
            res = "LABEL"
        elif self.VOLUME_VALUE.matches(word):
            res = int(self.VOLUME_VALUE.parseString(word)["value"][0])
        return res


class CityPrefixCategory(EnumCategory):
    '''
    I am the category for countries
    '''

    def __init__(self):
        '''
        constructor
        '''
        super().__init__("cityPrefix", lookupName="cityNamePrefixes")


class CityCategory(EnumCategory):
    '''
    I am the category for countries
    '''

    def __init__(self):
        '''
        constructor
        '''
        super().__init__("city", lookupName="cityNames")
        self.itemFunc=None

    def checkMatchWithContext(self, tokenStr: str, pos: int, tokenSequence) -> bool:
        fullname = self.getTokenStrWithPrefixes(tokenStr, pos, tokenSequence)
        matches = self.checkMatch(fullname)
        return matches

    def getValue(self, word: str, pos: int, tokenSequence):
        fullname = self.getTokenStrWithPrefixes(word, pos, tokenSequence)
        value = self.lookup(fullname)
        return value

    def getTokenStrWithPrefixes(self, tokenStr: str, pos:int, tokenSequence) -> str:
        tokenStr = tokenStr.strip(",;")
        cityPrefixTokens = [token for token in tokenSequence.matchResults if
                            isinstance(token.category, CityPrefixCategory)]
        cityPrefixTokens = [token for token in cityPrefixTokens if pos in [token.pos + prefixPos for prefixPos in
                                                                           [int(pos) for pos in
                                                                            token.value.split(",")]]]
        cityPrefixTokens.sort(key=lambda token: token.pos)
        fullname = " ".join([token.tokenStr for token in cityPrefixTokens]) + " " + tokenStr
        return fullname.strip()

class AcronymCategory(EnumCategory):
    '''
    I am the category for acronyms
    '''

    def __init__(self):
        '''
        constructor
        '''
        super().__init__("acronym", lookupName="acronyms")

    @staticmethod
    def sanitize(word:str):
        if isinstance(word, str):
            return word.strip(" ()#")

    def checkMatch(self, word):
        return super().checkMatch(self.sanitize(word))

    def lookup(self, word):
        return super().lookup(self.sanitize(word))


'''
Created on 2022-03-06

@author: wf
'''
from datetime import datetime
from corpus.utils.progress import Progress
from geograpy.locator import City
from lodstorage.tabulateCounter import TabulateCounter
from collections import Counter
from geograpy.locator import Locator
from corpus.location import LocationLookup
from corpus.lookup import CorpusLookup

class LocationFixer(object):
    '''
    fixer for locations - based on natural language location references
    such as 'Santa Fe, New Mexico, United States' the corresponding city, region
    and country information is looked up as a reference to wikiData and then
    added as the columns:
    city, region country and cityWikidataid, regionWikidataid and countryWikidataid
    '''
    
    def __init__(self):
        '''
        constructor
        '''
        # makes sure the geograpy3 data is available
        Locator.resetInstance()
        locator=Locator.getInstance()
        locator.downloadDB()
        self.locationLookup=LocationLookup()      
    
    def getCounter(self,events:list,propertyName:str):
        '''
        get a counter for the given event list and propertyName
        
        Args:
            events(list): a list of events to get counters for
            propertyName(str): the propertyname to count
            
        Returns:
            tuple(Counter,TabulateCounter): two counter representations
        '''
        counter=Counter()
        for event in events:
            if hasattr(event,propertyName):
                value=getattr(event,propertyName)
                if value is not None:
                    counter[value]+=1
        tabCounter=TabulateCounter(counter)
        return counter,tabCounter
    
    def timeStamp(self):
        '''
        get a timestamp with second precision
        
        Returns:
            str: an ISO-Date timestamp
        '''
        return datetime.utcnow().strftime('%Y-%m-%d_%H%M%S')
    
    def fixLocations4LookupIds(self,lookupIds,perCentLimit:float=60.0,logFileRoot="/tmp"):
        '''
        fix locations 4 the given lookup ids
        '''
        self.lookup=CorpusLookup(lookupIds=lookupIds)
        self.lookup.load(forceUpdate=False)
        for lookupId in lookupIds:
            dataSource=self.lookup.getDataSource(lookupId)
            eventManager=dataSource.eventManager
            show=True
            locationAttribute=dataSource.sourceConfig.locationAttribute
            limit=len(eventManager.getList())
            addLocationInfo=True
            showProgress=True
            logFile=None
            if logFileRoot is not None:
                timestamp=self.timeStamp()
                logFilePath=f"{logFileRoot}/{lookupId}-{timestamp}.log"
                logFile = open(logFilePath,'w')
            if lookupId=="dblp":
                eventManager.addLocations()
            self.fixLocations(eventManager, locationAttribute=locationAttribute, perCentLimit=perCentLimit,addLocationInfo=addLocationInfo, limit=limit, showProgress=showProgress,logFile=logFile)

    def fixLocations(self,eventManager,locationAttribute,perCentLimit=60.0,addLocationInfo=False,limit:int=100,showProgress=True,logFile=None):
        '''
        fix locations
        
        Args:
            eventManager: the eventmanager to use
            locationAttribute(str): the name of the location attribute
            addLocationInfo(bool): if True add the location information
            perCentLimit(float): percentage of Locations to cover (if less than 100% the fixing process will prematurely end at this limit)
        '''
        events=eventManager.events
        pCount,_pCountTab=self.getCounter(events,locationAttribute)
        eventsByLocation=eventManager.getLookup(locationAttribute,withDuplicates=True)
        count=len(pCount.items())
        total=sum(pCount.values())
        rsum=0
        problems=[]
        logHint=""
        if logFile:
            logHint=f"\nlogging to {logFile.name}"
        progress=Progress(progressSteps=1,expectedTotal=total,msg=f"Fixing {limit}/{total} locations for {eventManager.name} with perCentLimit {perCentLimit:.1f}% using location attribute {locationAttribute}{logHint}",showMemory=True)
        for i,locationTuple in enumerate(pCount.most_common(limit)):
            locationText,locationCount=locationTuple
            rsum+=locationCount
            percent=rsum/total*100
            city=None
            try:
                city=self.locationLookup.lookup(locationText,logFile=logFile)
            except Exception as ex:
                print(str(ex))
            if city is not None and isinstance(city,City):
                if showProgress:
                    progress.next()
                if logFile:
                    statsLine=f"{i:4d}/{count:4d}{rsum:6d}/{total:5d}({percent:4.1f}%)✅:{locationText}({locationCount})→{city} ({city.pop})"
                    print(statsLine,file=logFile,flush=True)
                events=eventsByLocation[locationText]
                # loop over all events and add location details
                for event in events:
                    event.city=city.name
                    event.cityWikidataid=city.wikidataid
                    
                    event.region=city.region.name
                    event.regionIso=city.region.iso
                    event.regionWikidataid=city.region.wikidataid
                    
                    event.country=city.country.name
                    event.countryIso=city.country.iso
                    event.countryWikidataid=city.country.wikidataid
            else:
                if logFile:
                    print(f"{i:4d}/{count:4d}{rsum:6d}/{total:5d}({percent:4.1f}%)❌:{locationText}({locationCount})",file=logFile)
                problems.append(locationText)
            if percent>perCentLimit:
                break
        problemMsg=f"found {len(problems)} problems"
        progress.done()
        for i,problem in enumerate(problems):
            if logFile:
                print(f"{i:4d}:{problem}",file=logFile)        
        print(problemMsg,file=logFile,flush=True)      
        if addLocationInfo:
            eventManager.store()
            print(f"{len(eventManager.events)} events stored",file=logFile,flush=True)
        


'''
Created on 2022-03-05

@author: wf
'''
import time
from corpus.utils.download import Profiler
import getpass
import os
import sys
import psutil

class Progress(object):
    '''
    Progress Display
    '''

    def __init__(self,progressSteps:int=None,expectedTotal:int=None,showDots:bool=False,msg:str=None,showMemory=False):
        '''
        Constructor
        
        Args:
            progressSteps(int): show progress after each number of the given number steps
            expectedTotal(int): expectedTotal
            showDots(boolean): if True show dots (e.g. for log files) else show a proper progress bar
            msg(str): message to display (if any)
            showMemory(bool): show memory usage
        '''
        self.count=0
        self.progressSteps=progressSteps
        self.expectedTotal=expectedTotal
        self.profiler=Profiler(msg=msg)
        self.startTime=self.profiler.starttime
        self.showMemory=showMemory
        #https://stackoverflow.com/a/63742743/1497139
        inUnitTest='unittest' in sys.modules
        self.showDots=showDots or inUnitTest #self.inCI()
        
    def inCI(self):
        '''
        are we running in a Continuous Integration Environment?
        '''
        publicCI=getpass.getuser() in ["travis", "runner"] 
        jenkins= "JENKINS_HOME" in os.environ
        return publicCI or jenkins

    def usedMemory(self):
        process = psutil.Process(os.getpid())
        mBytes=(process.memory_info().rss)/1024/1024  # in mbytes 
        return mBytes
  
    def printProgressBar (self,iteration, total, prefix = '', suffix = '', decimals = 1, length = 72, fill = '█', printEnd = "\r",startTime=None):
        """
        Call in a loop to create terminal progress bar
        
        see https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
        
        Args:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            decimals    - Optional  : positive number of decimals in percent complete (Int)
            length      - Optional  : character length of bar (Int)
            fill        - Optional  : bar fill character (Str)
            printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
        """
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + '-' * (length - filledLength)
        if startTime is None:
            print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
        else:
            elapsed=time.time()-startTime
            totalTime=elapsed*float(total)/iteration
            if self.showMemory:
                mbytes=self.usedMemory()
                suffix=f"{mbytes:.0f} MB"
            print(f'\r{prefix} |{bar}| {percent}% {elapsed:3.0f}/{totalTime:3.0f}s {suffix}', end = printEnd)
            
        # Print New Line on Complete
        if iteration == total: 
            print()
            
    def done(self,msg=None):
        '''
        progress is done
        
        Args:
            msg(str): an extra message to display
        '''
        if self.progressSteps is not None:
            if self.showDots:
                print("!")
            else:
                self.printProgressBar(self.expectedTotal, self.expectedTotal,startTime=self.startTime)
        if msg is None:
            msg=""
        self.profiler.time(extraMsg=msg)
        
    def next(self):
        '''
        count the progress and show it
        '''
        self.count+=1
        if self.progressSteps is not None:
            if self.showDots:   
                if self.count%self.progressSteps==0:
                    print(".",flush=True,end='')
                if self.count%(self.progressSteps*80)==0:
                    print(f"\n{self.count}",flush=True)
            else:
                if self.count%self.progressSteps==0:
                    self.printProgressBar(self.count, self.expectedTotal,startTime=self.startTime)


'''
Created on 2022-04-03

@author: wf
'''
import re
import datetime

class Textparse(object):
    '''
    Basic Natural Language Processing on Text
    '''

    def __init__(self):
        '''
        Constructor
        '''
        
    @staticmethod
    def strToDate(dateStr,dateFormat="%d.%m.%Y",debug:bool=False):
        '''
        convert the given string to a date
        
        Args:
            dateStr(str): the date string to convert
            dateFormat(str): the format to use for conversion
            debug(bool): if True show debug information
            
        Return:
            Date: - the converted data but None if there is a ValueError on conversion
        '''
        result=None
        try:
            result=datetime.datetime.strptime(
                        dateStr, dateFormat).date()
        except ValueError as ve:
            # TODO year might be ok but not garbage such as
            if debug:
                print(f"{dateStr}:{str(ve)}")
            pass
        return result
    
    @staticmethod
    def setDateRange(event,dateRange):
        for field in ["year","startDate","endDate"]:
            if field in dateRange:
                setattr(event,field,dateRange[field])
                
    @staticmethod
    def getDateRange(date):
        '''
        given a GND date string create a date range
        
        Args:
            date(str): the date string to analyze
            
        Returns:
            dict: containing year, startDate, endDate
        examples:
        2018-2019
        08.01.2019-11.01.2019
        2019
        1996.05.23-25
        2021.02.28-03.02
        '''
        result={}
        if date is not None:
            startDate=None
            endDate=None
            yearPattern="[12][0-9]{3}"
            monthPattern="01|02|03|04|05|06|07|08|09|10|11|12"
            dayPattern="01|02|03|04|05|06|07|08|09|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31"
            datePattern="[0-9]{2}[.][0-9]{2}[.]"+yearPattern
            yearOnly=re.search(r"^("+yearPattern+")[-]?("+yearPattern+")?$",date)
            if yearOnly:
                result['year']=int(yearOnly.group(1))
            else:
                fromOnly=re.search(r"^("+datePattern+")[-]?$",date)
                if fromOnly:
                    startDate=Textparse.strToDate(fromOnly.group(1))
                else:
                    fromTo=re.search(r"^("+datePattern+")[-]("+datePattern+")$",date)
                    if fromTo:
                        startDate=Textparse.strToDate(fromTo.group(1))
                        endDate=Textparse.strToDate(fromTo.group(2))
                    else:
                        fromToSimple=re.search(r"^(?P<year>"+yearPattern+")\.(?P<month>"+monthPattern+")\.(?P<fromday>"+dayPattern+")[-]"+
                                               "((?P<untilmonth>"+monthPattern+")\.(?P<untildayWithMonth>"+dayPattern+")|(?P<untilday>"+dayPattern+"))$",date)
                        if fromToSimple:
                            year=fromToSimple.group("year")
                            month=fromToSimple.group("month")
                            untilmonth=fromToSimple.group("untilmonth")
                            if untilmonth is None:
                                untilmonth=month
                                untilday=fromToSimple.group("untilday")
                            else:
                                untilday=fromToSimple.group("untildayWithMonth")
                            fromday=fromToSimple.group("fromday")
                            dateFormat="%Y-%m-%d"
                            startDate=Textparse.strToDate(f"{year}-{month}-{fromday}",dateFormat=dateFormat)
                            endDate=Textparse.strToDate(f"{year}-{untilmonth}-{untilday}",dateFormat=dateFormat)
                            pass
            if startDate:
                result['startDate']=startDate
            if endDate:
                result['endDate']=endDate
        if 'startDate' in result:
                result['year']=result['startDate'].year
        return result

        


'''
Created on 2021-11-14

@author: wf
'''
from corpus.lookup import CorpusLookup
from corpus.datasources.openresearch import OREvent

class EventExporter():
    '''
    exporter for Events and series
    '''
    
    def __init__(self,debug:bool=False):
        '''
        construct me
        '''
        self.debug=debug
        lookup=CorpusLookup(lookupIds=["dblp","wikidata","confref"])
        lookup.load(forceUpdate=False)
        self.dblpDataSource=lookup.getDataSource("dblp")
        self.confrefDataSource=lookup.getDataSource("confref")
        wikidataDataSource=lookup.getDataSource("wikidata")
        self.wikiDataSeriesByDblpPid,_dup=wikidataDataSource.eventSeriesManager.getLookup("DBLP_pid")
        self.dblpSeriesById,_dup=self.dblpDataSource.eventSeriesManager.getLookup("eventSeriesId")
        self.orTemplateParamLookup=OREvent.getTemplateParamLookup()
    
    def exportSeries2OpenResearch(self,dblpSeriesId):
        '''
        export the series with the given dblp Series Id to OpenResearch
        
        Args:
            dblpSeriesId(str): the id of the dblp series to be exported
            
        Return:
            int: the number of events exported
        '''
        count=0
        if dblpSeriesId in self.dblpSeriesById:
            eventSeries=self.dblpSeriesById[dblpSeriesId]
            eventBySeries=self.dblpDataSource.eventManager.getLookup("series",withDuplicates=True)
            events=eventBySeries[dblpSeriesId]
            count=self.exportSeries2OpenRessearch(eventSeries, events)
        return count
            
    def exportSeries2OpenRessearch(self,eventSeries,events):
        '''
        Args:
            eventSeries(EventSeries): the series to export
            events(list): the list of events to export
        Return:
            int: the number of events exported
        '''
        count=0
        for event in events:
            pass
        markup=eventSeries.asWikiMarkup()
        print(markup)
        for event in events:
            markup=event.asWikiMarkup(eventSeries.acronym,self.orTemplateParamLookup)
            print(markup)
            count+=1
        return count


'''
Created on 2021-11-04

@author: wf
'''
from corpus.event import Event,EventSeries,EventManager,EventSeriesManager
from lodstorage.storageconfig import StorageConfig
from corpus.eventcorpus import EventDataSourceConfig,EventDataSource
from corpus.datasources.webscrape import WebScrape
from corpus.datasources.wikidata import Wikidata

class ACM(EventDataSource):
    '''
    ACM digital library access
    https://dl.acm.org/conferences
    '''
    sourceConfig=EventDataSourceConfig(lookupId="acm",name="acm",url="https://dl.acm.org/conferences",title="Association for Computing machinery",tableSuffix="acm")
  

    def __init__(self,debug:bool=False,showHtml=False):
        '''
        Constructor
        
            debug(bool): True if debugging should be active
            showHtml(bool): True if HTML of scraped websites should be shown
        '''
        super().__init__(AcmEventManager(),AcmEventSeriesManager(),ACM.sourceConfig)
        self.debug=debug
        self.showHtml=showHtml
        
    def getSoup(self,url:str):
        '''
        get the beautiful soup parser result for the given url
        
        Args:
            url(str): the url to parse
            
        Return:
            BeautifulSoup: the soup
        '''
        msg=f"getting {url} ..."
        if self.debug:
            print (msg)
        scrape=WebScrape(debug=self.debug)
        soup=scrape.getSoup(url, showHtml=self.showHtml)   
        return soup
    
    def eventFromProceedingsDOI(self,doi:str):
        '''
        create an AcmEvent from the given ACM Digital Library Proceedings doi
        
        Args:
            acmDlEventId(doi): ACM Digital Library Proceedings DOI
           
        Return:
            AcmEvent
            
        '''
        url=f"https://dl.acm.org/doi/proceedings/{doi}"
        eventSoup=self.getSoup(url)
        acmEvent=self.eventFromSoup(eventSoup)
        return acmEvent
    
    def eventSeriesfromDigitalLibraryEventId(self,acmDlEventId:str):
        '''
        create an AcmEventSeries from the given ACM Digital Library Event Id
        
        Args:
            acmDlEventId(str): ACM Digital Library Event Id see https://www.wikidata.org/wiki/Property:P3333
        '''
        url=f"https://dl.acm.org/event.cfm?id={acmDlEventId}"
        seriesSoup=self.getSoup(url)
        canonical = seriesSoup.find('link', {'rel': 'canonical'})
        link=canonical['href']
        proceedingsUrl=f"{link}/proceedings"
        proceedingsSoup=self.getSoup(proceedingsUrl)
        acmSeries=self.eventSeriesFromSoups(seriesSoup,proceedingsSoup)
        return acmSeries
        
    def eventSeriesFromSoups(self,seriesSoup,proceedingsSoup,debug=False):
        '''
        create an eventSeries from two beautiful soup parsed html pages
        '''
        acmSeries=AcmEventSeries()
        #debug=True
        if debug:
            seriesHTML = seriesSoup.prettify()
            print (seriesHTML)
        
        return acmSeries
    
    def eventFromSoup(self,eventSoup):
        '''
        create an event from the parsed html page
        '''
        acmEvent=AcmEvent()
        return acmEvent
        
class AcmEvent(Event):
    '''
    a scientific event derived from the ACM digital library
    '''
    
    
class AcmEventSeries(EventSeries):
    '''
    a scientific event series derived from the ACM digital library
    '''
          
class AcmEventManager(EventManager):
    '''
    ACM event manager
    '''
        
    def __init__(self, config: StorageConfig = None):
        '''
        Constructor
        '''
        self.source="acm"
        self.endpoint=Wikidata.endpoint
        super().__init__(name="ACMEvents", sourceConfig=ACM.sourceConfig, clazz=AcmEvent, config=config)

    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts=self.getLoDfromEndpoint    
   
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this event manager 
        '''
        query="""# WF 2021-11-12
# ACM proceedings
SELECT ?doi ?publisher ?publisherLabel ?proc ?procLabel ?event ?eventLabel ?type ?typeLabel ?series ?seriesLabel ?acmConferenceId ?acmEventSeriesId ?gndEventId ?dblpEventSeriesId WHERE {
  # proceedings
  ?proc wdt:P31 wd:Q1143604.
  # publisher ACM
  ?proc wdt:P123 wd:Q127992.
  OPTIONAL {
    ?proc wdt:P123 ?publisher
  }
  # doi
  OPTIONAL {
    ?proc wdt:P356 ?doi.
  }
  OPTIONAL {
    # of an event
    ?proc wdt:P4745 ?event.
    OPTIONAL {
      ?event wdt:P227 ?gndEventId.
    }
    OPTIONAL {
      ?event wdt:P31 ?type.
    } 
    OPTIONAL {
      ?event wdt:P179 ?series
  
      OPTIONAL {
        ?series wdt:P7979 ?acmConferenceId.
      }
      OPTIONAL {
        ?series wdt:P3333 ?acmEventSeriesId.
      }
      OPTIONAL  {
       ?series wdt:P8926 ?dblpEventSeriesId.
      } 
    }
  }
  SERVICE wikibase:label { bd:serviceParam wikibase:language "en". }
}
        """
        return query      
    
class AcmEventSeriesManager(EventSeriesManager):
    '''
    event series handling for event series derived from the ACM digital library
    '''
     
    def __init__(self,config:StorageConfig=None):
        '''
        Constructor
        '''
        self.source="acm"
        self.endpoint=Wikidata.endpoint
        super().__init__(name="AcmEventSeries", sourceConfig=ACM.sourceConfig,clazz=AcmEventSeries,config=config)

    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts=self.getLoDfromEndpoint
            
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this series  manager
        '''
        query="""# WF 2021-11-04
# ACM event series in Wikidata
SELECT ?eventSeries ?eventSeriesLabel ?acmConferenceId ?acmEventId ?dblpEventId  WHERE {
  # scientific conference series
  ?eventSeries wdt:P31 wd:Q47258130.  
  #?eventSeries wdt:P31 ?type.
  ?eventSeries wdt:P7979 ?acmConferenceId.
  ?eventSeries wdt:P3333 ?acmEventId.
  OPTIONAL  {
    ?eventSeries wdt:P8926 ?dblpEventId.
  }
  SERVICE wikibase:label { bd:serviceParam wikibase:language "en". }
}
ORDER by ?acmEventId"""
        return query


'''
Created on 2021-07-21

@author: wf
'''
from corpus.event import EventSeriesManager,EventSeries, Event, EventManager, EventStorage
from lodstorage.storageconfig import StorageConfig
from corpus.eventcorpus import EventDataSource,EventDataSourceConfig

class Wikidata(EventDataSource):
    '''
    Wikidata access via SPARQL endpoint
    
    make do not want to be dependend on this endpoint since we might have
    our own copy of Wikidata which might run on Virtuoso or Jena instead of blazegraph
    '''
    endpoint="https://query.wikidata.org/sparql"
    sourceConfig=EventDataSourceConfig(lookupId="wikidata",name="Wikidata",url='https://www.wikidata.org/wiki/Wikidata:Main_Page',title='Wikidata',tableSuffix="wikidata",locationAttribute="location")
    
    def __init__(self):
        '''
        construct me
        '''
        super().__init__(WikidataEventManager(),WikidataEventSeriesManager(),Wikidata.sourceConfig)
        
class WikidataEventSeries(EventSeries):
    '''
    event series derived from Wikidata
    '''
    
    @staticmethod
    def postProcessLodRecord(rawEvent:dict):
        '''
        fix the given raw EventSeries
        
        Args:
            rawEvent(dict): the raw event record to fix
        '''
        rawEvent["source"]="wikidata"
        for key in ["eventSeriesId"]:
            if key in rawEvent:
                rawEvent["url"]=rawEvent[key]
                rawEvent[key]=rawEvent[key].replace("http://www.wikidata.org/entity/","")
    
class WikidataEvent(Event):
    '''
    event derived from Wikidata
    '''
    
    @staticmethod
    def postProcessLodRecord(rawEvent:dict):
        '''
        fix the given raw Event
        
        Args:
            rawEvent(dict): the raw event record to fix
        '''
        rawEvent["source"]="wikidata"
        for key in ["eventId","countryId","locationId","eventInSeriesId"]:
            if key in rawEvent:
                rawEvent[key]=rawEvent[key].replace("http://www.wikidata.org/entity/","")
        if 'startDate' in rawEvent:
            startDate=rawEvent['startDate']
            if startDate:
                rawEvent['year']=startDate.year
                pass
        if 'ordinal' in rawEvent:
            try:
                rawEvent['ordinal']=int(rawEvent["ordinal"])
            except Exception as _ex:
                # ignore the invalid ordinal
                del(rawEvent['ordinal'])
                
    
class WikidataEventManager(EventManager):
    '''
    manage wikidata derived scientific events
    '''
    def __init__(self,config:StorageConfig=None):
        '''
        Constructor
        '''
        self.source="wikidata"
        self.endpoint=Wikidata.endpoint
        self.queryManager=EventStorage.getQueryManager(lang="sparql",name="wikidata")
        super(WikidataEventManager,self).__init__(name="WikidataEvents", sourceConfig=Wikidata.sourceConfig,clazz=WikidataEvent,config=config)
        
    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts=self.getLoDfromEndpoint    
   
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this event manager
        
        see also 
           https://github.com/TIBHannover/confiDent-dataimports/blob/master/wip/wikidata_academic_conferences.rq
           https://confident.dbis.rwth-aachen.de/or/index.php?title=Iteration1_Property_Mapping
        '''
        eventQuery=self.queryManager.queriesByName['Wikidata-Events']
        query=eventQuery.query
        return query
    
class WikidataEventSeriesManager(EventSeriesManager):
    '''
    wikidata scientific conference Series Manager
    '''

    def __init__(self,config:StorageConfig=None):
        '''
        Constructor
        '''
        self.source="wikidata"
        self.endpoint=Wikidata.endpoint
        self.queryManager=EventStorage.getQueryManager(lang="sparql",name="wikidata")
        super(WikidataEventSeriesManager,self).__init__(name="WikidataEventSeries", sourceConfig=Wikidata.sourceConfig,clazz=WikidataEventSeries,config=config)
        
    def configure(self):
        '''
        configure me
        '''
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts=self.getLoDfromEndpoint
        
    def getSparqlQuery(self):
        '''
        get  the SPARQL query for this series  manager
        '''
        eventQuery=self.queryManager.queriesByName['Wikidata-Eventseries']
        query=eventQuery.query
        return query
        
        


'''
Created on 2021-08-11

@author: wf
'''
#from lodstorage.entity import EntityManager
from geograpy.locator import LocationContext
from corpus.nominatim import NominatimWrapper
import sys
class LocationLookup:
    '''
    lookup locations
    '''
    preDefinedLocations={
        "Not Known": None,
        "Online": None,
        "Virtual Event, USA": None,
        "Albuquerque, New Mexico, USA":"Q34804",
        "Alexandria, Virginia, USA":"Q88",
        "Alexandria, VA": "Q88",
        "Amsterdam": "Q727",
        "Amsterdam, Amsterdam": "Q727",
        "Amsterdam Netherlands": "Q727",
        "Amsterdam, Netherlands": "Q727",
        "Amsterdam, The Netherlands":"Q727",
        "Bergen, Norway":"Q26793",
        "Bremen, Germany": "Q24879",
        "Brussels, Belgium": "Q239",
        "Cancun, Mexico":"Q8969",
        "Cancún, Mexico": "Q8969",
        "Cambridge, United Kingdom": "Q21713103",
        "Cambridge, UK": "Q21713103",
        "Cambridge, USA": "Q49111",
        "Cambridge, MA":"Q49111",
        "Cambridge, Massachusetts": "Q49111",
        "Cambridge, Massachusetts, USA":"Q49111",
        "Cambridge, MA, USA":"Q49111",
        "Charleston, South Carolina, USA":"Q47716",
        "Gdansk, Poland":"Q1792",
        "Heraklion, Crete, Greece":"Q160544",
        "Los Angeles California": "Q65",
        "Los Angeles CA USA": "Q65",
        "Luxembourg, Luxembourg":"Q1842",
        "Macau, Macau, China":"Q14773",
        "Marina del Rey, CA": "Q988140",
        "Monterrey, Mexico":"Q81033",
        "Montreal, QC": "Q340",
        "Montréal, QC": "Q340",
        "Montreal, QC, Canada": "Q340",
        "Montrèal, Canada": "Q340",
        "New Brunswick, New Jersey, USA":"Q138338",
        "New Brunswick, New Jersey": "Q138338",
        "New Delhi": "Q987",
        "New Delhi, India": "Q987",
        "New Orleans, LA": "Q34404",
        "Palo Alto, USA": "Q47265",
        "Palo Alto, CA": "Q47265",
        "Palo Alto, California": "Q47265",
        "Palo Alto, California, USA": "Q47265",
        "Pasadena, California, USA":"Q485176",
        "Phoenix": "Q16556",
        "Phoenix, AZ": "Q16556",
        "Phoenix, Arizona": "Q16556",
        "Phoenix AZ USA": "Q16556",
        "Phoenix, Arizona, USA": "Q16556",
        "Phoenix, USA": "Q16556",
        "Phoenix, AZ, USA": "Q16556",
        "Salamanca, Spain": "Q15695",
        "Santa Barbara, California": "Q159288",
        "Santa Barbara, CA": "Q159288",
        "Santa Barbara, CA, USA": "Q159288",
        "Santa Barbara CA USA":  "Q159288",
        "Santa Barbara, USA": "Q159288",
        "Santa Barbara, California, USA": "Q159288",
        "Santa Fe, New Mexico": "Q38555",
        "Santa Fe, NM, USA": "Q38555",
        "Santa Fe, New Mexico, USA": "Q38555",
        "Santa Fe, USA": "Q38555",
        "Santa Fe, New Mexico, United States": "Q38555",
        "Skovde, Sweden": "Q21166",
        "Snowbird, Utah, USA": "Q3487194",
        "St. Louis, MO, USA": "Q38022",
        "St. Petersburg": "Q656",
        "Saint-Petersburg, Russia":"Q656",
        "Thessaloniki": "Q17151",
        "Thessaloniki, Greece": "Q17151",
        "Toronto, ON": "Q172",
        "Toronto, Ontario": "Q172",
        "Toronto, Ontario, Canada": "Q172",
        "Toronto, Canada": "Q172",
        "Trondheim, Norway":"Q25804",
        "Valencia": "Q8818",
        "Valencia, Spain": "Q8818",  
        "Valencia, Valencia, Spain": "Q8818",  
        "York, UK":"Q42462"
    }
    other={
        "Washington, DC, USA": "Q61",
        "Bangalore": "Q1355",
        "Bangalore, India": "Q1355",
        "Xi'an": "Q5826",
        "Xi'an, China": "Q5826",
        "Virtual Event USA": "Q30",
        "Virtual USA": "Q30",
        "London United Kingdom": "Q84",
        "Brno":"Q14960",
        "Cancun":"Q8969",
        "Gothenburg Sweden": "Q25287",
        
        "Zurich, Switzerland": "Q72",
        "Barcelona Spain": "Q1492",
        "Vienna Austria": "Q1741",
        "Seoul Republic of Korea": "Q8684",
        "Seattle WA USA": "Q5083",
        "Singapore Singapore":"Q334",
        "Tokyo Japan": "Q1490",
        "Vancouver BC Canada": "Q24639",
        "Vancouver British Columbia Canada": "Q24639",
        "Paris France": "Q90",
        "Nagoya": "Q11751",
        "Marrakech":"Q101625",
        "Austin Texas":"Q16559",
        "Chicago IL USA":"Q1297",
        "Bangkok Thailand":"Q1861",
        "Firenze, Italy":"Q2044",
        "Florence Italy":"Q2044",
        "Timisoara":"Q83404",
        "Langkawi":"Q273303",
        "Beijing China":"Q956",
        "Berlin Germany": "Q64",
        "Prague Czech Republic":"Q1085",
        "Portland Oregon USA":"Q6106",
        "Portland OR USA":"Q6106",
        "Pittsburgh PA USA":"Q1342",
        "Новосибирск":"Q883",
        "Los Angeles CA USA":"Q65",
        "Kyoto Japan": "Q34600"
    }

    def __init__(self):
        '''
        Constructor
        '''
        self.locationContext=LocationContext.fromCache()
        cacheRootDir=LocationContext.getDefaultConfig().cacheRootDir
        cacheDir=f"{cacheRootDir}/.nominatim"
        self.nominatimWrapper=NominatimWrapper(cacheDir=cacheDir)
        
    def getCityByWikiDataId(self,wikidataID:str):
        '''
        get the city for the given wikidataID
        '''
        citiesGen=self.locationContext.cityManager.getLocationsByWikidataId(wikidataID)
        if citiesGen is not None:
            cities=list(citiesGen)
            if len(cities)>0:
                return cities[0]
        else:
            return None
        
        
    def lookupNominatim(self,locationText:str):
        '''
        lookup the location for the given locationText (if any)
        
        Args:
            locationText(str): the location text to search for
            
        Return:
            the location by first finding the wikidata id for the location text and then looking up the location
        '''
        location=None
        wikidataId=self.nominatimWrapper.lookupWikiDataId(locationText)
        if wikidataId is not None:
            location=self.getCityByWikiDataId(wikidataId)
        return location
        
    def lookup(self,locationText:str,logFile=sys.stdout):
        '''
        lookup a location based on the given locationText
        
        Args:
            locationText(str): the location to lookup
            logFile: the logFile to use - default is sys.stdout
        '''
        if locationText in LocationLookup.preDefinedLocations:
            locationId=LocationLookup.preDefinedLocations[locationText]
            if locationId is None:
                return None
            else:
                location=self.getCityByWikiDataId(locationId)
                if location is None:
                    print(f"❌❌-predefinedLocation {locationText}→{locationId} wikidataId not resolved",file=logFile)
                return location
        lg=self.lookupGeograpy(locationText)
        ln=self.lookupNominatim(locationText)
        if ln is not None and lg is not None and not ln.wikidataid==lg.wikidataid:
            print(f"❌❌{locationText}→{lg}!={ln}",file=logFile)
            return None
        return lg
        
    def lookupGeograpy(self,locationText:str):
        '''
        lookup the given location by the given locationText
        '''
        locations=self.locationContext.locateLocation(locationText)
        if len(locations)>0:
            return locations[0]
        else:
            return None
        


'''
Created on 2022-05-17

@author: wf
'''
import io
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np
from typing import Callable, List
from collections import Counter

import pandas as pd
from scipy.optimize import minimize
from scipy.optimize import curve_fit
from scipy.special import zetac

class PlotSettings():
    '''
    plot settings
    '''
    def __init__(self,imageFormat='png',interactive:bool=False,outputFile:str=None,callback:Callable=None):
        '''
        Constructor 
        
        Args:
            imageFormat(str): the format for the outputFile
            interactive(bool): if True show using matplotlib backend
            outputFile(str): the path to the outputfile
            callback(Callable): optional function to call to set more details of figure/plot
        '''
        self.imageFormat=imageFormat
        self.interactive=interactive
        self.outputFile=outputFile
        self.callback=callback
        
            
class Plot():
    '''
    super class of all plots
    '''
    def __init__(self, **kwargs):
        pass
    
    def setup(self,title:str):
        '''
        setup my figure and  plot (is global anyway but shouldn't be ...)
        
        Args:
            title(str)
        '''
        self.fig = plt.figure(title)
        self.plt = plt
        # clear
        plt.clf()
        plt.title(title)
    
    def doShow(self,ps:PlotSettings=None):
        '''
        '''
        if not ps:
            return
        if ps.callback:
            ps.callback(self)
        if ps.interactive and not ps.outputFile:
            self.plt.show()
            return None
        else:
            if ps.outputFile is None:
                ps.outputFile = io.BytesIO()
            self.plt.savefig(ps.outputFile, format=ps.imageFormat,dpi=self.fig.dpi)
            plt.close(self.fig)
            return ps.outputFile

class Zipf(Plot):
    '''
    
    Powerlaw/Zipf Distribution plot and analysis
    
    https://stats.stackexchange.com/questions/331219/characterizing-fitting-word-count-data-into-zipf-power-law-lognormal
    https://stats.stackexchange.com/questions/6780/how-to-calculate-zipfs-law-coefficient-from-a-set-of-top-frequencies
    https://stackoverflow.com/questions/12037494/curve-fitting-zipf-distribution-matplotlib-python
    
    '''
    def __init__(self, values:list,minIndex:int=0):
        '''
        Constructor
        '''
        self.values=values
        self.counter_of_values = Counter(self.values)
        xValues=sorted(list(self.counter_of_values.keys()))
        xValues=xValues[minIndex:]
        self.x=np.array(xValues)
        yValues=[self.counter_of_values[x] for x in xValues]
        self.freqs=np.array(yValues) 
        self.mean = np.mean(self.freqs)
        self.freqsLog=np.log(self.freqs)
        self.freqsLogMean= np.mean(self.freqsLog)
        
    def curveFit(self):

        def f(x, a):
            return (x**-a)/zetac(a)


        result = curve_fit(f, self.x, self.freqs)
        return result

        
    def fit(self):
        '''
        calculate the zipf distribution fit
        '''
        def loglik(b):  
            # Power law function
            Probabilities = self.x**(-b)
        
            # Normalized
            Probabilities = Probabilities/Probabilities.sum()
        
            # Log Likelihoood
            # Lvector = np.log(Probabilities)
        
            # Multiply the vector by frequencies
            Lvector = np.log(Probabilities) * self.freqs
        
            # LL is the sum
            L = Lvector.sum()
        
            # We want to maximize LogLikelihood or minimize (-1)*LogLikelihood
            return(-L)

        fitResult=minimize(loglik, [2]) 
        return fitResult

        
    def show(self,title:str="Power law",ps:PlotSettings=None):
       
        #counter_of_counts = Counter(counter_of_values.values())
        #value_counts = np.array(list(counter_of_counts.keys()))
        #freq_of_value_counts = np.array(list(counter_of_counts.values()))
        #plt.scatter(np.log(value_counts), np.log(freq_of_value_counts))
        self.setup(title)
        plt.scatter(
            self.x,
            self.freqsLog,       
        )
        plt.xlabel('Values')
        plt.ylabel('Log of count')
        fit=self.fit()
        curveFit=self.curveFit()
        sx=1
        sy=self.freqsLogMean-1
        if fit.success:
            s=fit.x[0]
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
            popt,pcov=curveFit
            cfs=popt[0]
            #plt.text(sx, sy, fr'Zipf distribution fit s={s:.3f} cfs:{cfs:.3f} cov:{pcov}')
            #plt.plot(self.x, self.x*-s)
        else:
            plt.text(sx, sy, r'no Zipf distribution fit')
        self.doShow(ps)
        

class Histogramm(Plot):
    '''
    Histogram
    '''

    def __init__(self, x):
        self.x=x

    def prepareHistogram(self,facecolor='b',alpha=0.5,density:bool=False, bins=None, label=None):
        self.addHistogramSeries(x=self.x,
                                facecolor=facecolor,
                                alpha=alpha,
                                density=density,
                                bins=bins,
                                label=label)

    def addHistogramSeries(self, x, facecolor='b',alpha=0.5,density:bool=False, bins=None, label=None):
        if bins is None:
            bins = np.arange(min(x) - 0.5, max(x) + 1.5, 1.0)
        _n, _bins, _patches = plt.hist(x,
                                       bins,
                                       density=density,
                                       facecolor=facecolor,
                                       alpha=alpha,
                                       rwidth=0.9,
                                       label=label)
    
    def show(self,xLabel,yLabel,title,facecolor='b',alpha=0.5,density:bool=False,grid:bool=True,ps:PlotSettings=None,bins=None, vlineAt:int=None, label=None):
        '''
        show the histogram
        
        Args:
            xLabel(str): the x-axis label
            yLabel(str): the y-axis label
            title(str): the title
            facecolor(str): the facecolor to use
            alpha(float): how opaque
            density(bool): if True show the relative values
            grid(bool): if True show the histogramm grid
            ps(PlotSettings): plot settings
            
        '''
        # the histogram of the data
        self.setup(title)
        self.prepareHistogram(facecolor=facecolor,
                              alpha=alpha,
                              density=density,
                              bins=bins,
                              label=label)
        plt.xlabel(xLabel)
        plt.ylabel(yLabel)
        if density:
            yvals = plt.gca().get_yticks()
            #print(yvals)
            plt.gca().set_yticklabels([f"{y*100:.2f}%"for y in yvals])
        if vlineAt is not None:
            plt.gca().axvline(x=vlineAt, color='r', linestyle='dashed', linewidth=2)
        plt.grid(grid)
        plt.legend(loc='upper right')
        self.doShow(ps)


class HistogramSeries(Histogramm):
    '''
    Histogram of multiple Series
    '''

    def __init__(self, series:dict):
        if series is None:
            series = {}
        super().__init__(series)

    def addSeries(self, label:str, series):
        """
        Add a series to the histogram
        Args:
            label: label of the series (displayed in the legend)
            series: series data
        """
        self.x.update(label, series)

    def prepareHistogram(self, facecolor='b', alpha=0.5, density: bool = False, bins=None, label=None):
        """
        Prepares the histogram with the given parameters
        Args:
            facecolor:
            alpha:
            density:
            bins:
            label:

        Returns:

        """
        for i, (label, series) in enumerate(self.x.items()):
            color = list(mcolors.BASE_COLORS.keys())[i % len(mcolors.BASE_COLORS)]
            self.addHistogramSeries(x=series,
                                    facecolor=color,
                                    alpha=alpha,
                                    density=density,
                                    bins=bins,
                                    label=label)
        
            


'''
Created on 2021-07-26

@author: wf
'''
from typing import Callable
from corpus.config import EventDataSourceConfig
from lodstorage.csv import CSV
from lodstorage.entity import EntityManager
from lodstorage.jsonable import JSONAble
from lodstorage.lod import LOD
from lodstorage.sql import SQLDB
from corpus.utils.download import Profiler
from lodstorage.storageconfig import StorageConfig
from corpus.quality.rating import RatingManager
from corpus.eventrating import EventRating,EventSeriesRating
from lodstorage.sparql import SPARQL
from lodstorage.schema import Schema
from lodstorage.uml import UML
from lodstorage.query import QueryManager
import os
import sys
from datetime import datetime


import re
class EventStorage:
    '''
    common storage aspects of the EventManager and EventSeriesManager
    '''
    profile=True
    withShowProgress=False
    viewTableExcludes={
            "event":
                ["event_acm",
                 "event_ceurws",
                 "event_orclonebackup",
                 "event_or",
                 "event_orbackup"],
            "eventseries":
                ["eventseries_acm",
                 "eventseries_or",
                 "eventseries_orbackup",
                 "eventseries_orclonebackup",
                 "eventseries_gnd"]
            }
    
    @staticmethod
    def getStorageConfig(debug:bool=False,mode='sql')->StorageConfig:
        '''
        get the storageConfiguration
        
        Args:
            debug(bool): if True show debug information
            mode(str): sql or json
        
        Return:
            StorageConfig: the storage configuration to be used
        '''
        if mode=='sql':
            config=StorageConfig.getSQL(debug=debug)
        elif mode=='json':
            config=StorageConfig.getJSON()
        elif mode=='jsonpickle':
            config=StorageConfig.getJsonPickle(debug=debug)
        else:
            raise Exception(f"invalid mode {mode}")
        config.cacheDirName="conferencecorpus"
        cachedir=config.getCachePath() 
        config.profile=EventStorage.profile
        config.withShowProgress=EventStorage.withShowProgress
        if mode=='sql':
            config.cacheFile=f"{cachedir}/EventCorpus.db"
        return config
    
    @classmethod
    def getQueryManager(cls,lang='sql',name="queries",debug=False):
        '''
        get the query manager for the given language and fileName
        
        Args:
            lang(str): the language of the queries to extract
            name(str): the name of the manager containing the query specifications
            debug(bool): if True set debugging on
        '''
        cachedir=EventStorage.getStorageConfig().getCachePath()
        for path in cachedir,os.path.dirname(__file__)+"/../resources":
            qYamlFile=f"{path}/{name}.yaml"
            if os.path.isfile(qYamlFile):
                qm=QueryManager(lang=lang,debug=debug,queriesPath=qYamlFile)
                return qm
        return None
    
    @classmethod
    def getDBFile(cls,cacheFileName="EventCorpus"):
        '''
        get the database file for the given cacheFileName
        
        Args:
            cacheFileName(str): the name of the cacheFile without suffix
        '''
        config=cls.getStorageConfig()
        cachedir=config.getCachePath()
        dbfile=f"{cachedir}/{cacheFileName}.db"
        dbfile=os.path.abspath(dbfile)
        return dbfile
    
    @classmethod
    def getSqlDB(cls):
        '''
        get the SQL Database
        '''
        dbfile=EventStorage.getDBFile()
        sqlDB=SQLDB(dbfile)
        return sqlDB
    
    @classmethod
    def getTableList(cls,withInstanceCount:bool=True)->list:
        '''
        get the list of SQL Tables involved
        
        Return:
            list: the map of SQL tables used for caching
            withInstanceCount(bool): if TRUE add the count of instances to the table Map 
        '''
        sqlDB=EventStorage.getSqlDB()
        tableList=sqlDB.getTableList()
        for table in tableList:
            tableName=table["name"]
            if withInstanceCount:
                countQuery="SELECT count(*) as count from %s" % tableName
                countResult=sqlDB.query(countQuery)
                table['instances']=countResult[0]['count']
        return tableList
    
    @classmethod
    def getViewTableList(cls,viewName,exclude=None):
        sqlDB=EventStorage.getSqlDB()
        tableList=sqlDB.getTableList()  
        viewTableList=[]
        for table in tableList:
            tableName=table["name"]
            if tableName.startswith(f"{viewName}_"):
                if exclude is None or tableName not in exclude[viewName]:
                    viewTableList.append(table)
        return viewTableList
    
    @classmethod
    def getCommonViewDDLs(cls,viewNames=["event","eventseries"],exclude=None):
        '''
        get the SQL DDL for a common view 
        
        Return:
            str: the SQL DDL CREATE VIEW command
        '''
           
        viewDDLs={}
        for viewName in viewNames:
            viewTableList=cls.getViewTableList(viewName, exclude=exclude)
            viewDDL=Schema.getGeneralViewDDL(viewTableList, viewName)
            viewDDLs[viewName]=viewDDL
        return viewDDLs
        
    @classmethod
    def createViews(cls,exclude=None,show=False):
        ''' 
          create the general Event views
          
        Args:
            exclude(list): the list of table names to be excluded
            show(bool): if True show the DDL
        '''
        sqlDB=EventStorage.getSqlDB()
        viewDDLs=EventStorage.getCommonViewDDLs(exclude=exclude)
        for viewName,viewDDL in viewDDLs.items():
            sqlDB.c.execute(f"DROP VIEW IF EXISTS {viewName}")
            if show:
                print(viewDDL)
            sqlDB.c.execute(viewDDL)
    
    @classmethod        
    def asPlantUml(cls,baseEntity='Event',exclude=None):
        '''
        return me as a plantUml Diagram markup
        '''
        schemaManager=None
        uml=UML()
        now=datetime.now()
        nowYMD=now.strftime("%Y-%m-%d")
        viewName=f"{baseEntity.lower()}"
        tableList=EventStorage.getViewTableList(viewName, exclude=exclude)
        for table in tableList:
            tableName=table['name']
            if 'instances' in table:
                instanceNote=""
                dataSource=self.getDataSource4TableName(tableName)
                if dataSource is not None:
                    sourceConfig=dataSource.sourceConfig
                    instanceNote=f"[[{sourceConfig.url} {sourceConfig.title}]]"
                instanceCount=table['instances']
                instanceNote=f"{instanceNote}\n{instanceCount} instances "
                table['notes']=instanceNote
        title=f"""ConfIDent  {baseEntity}
{nowYMD}
[[https://projects.tib.eu/en/confident/ © 2019-2022 ConfIDent project and Wolfgang Fahl]]
see also [[http://cc.bitplan.com Conference Corpus]]
"""
        plantUml=uml.mergeSchema(schemaManager,tableList,title=title,packageName='DataSources',generalizeTo=baseEntity)
        return plantUml
    
    @classmethod
    def getSignatureCache(cls,profile:bool=True,force:bool=False):
        '''
        cache the signature Data in a separate SQLite DB
        
        Args:
            profile(bool): if True show profiling information
            force(bool): if True force the cache creation
        '''
        signatureCache=cls.getDBFile("Signature")
        profiler=None
        if (not os.path.isfile(signatureCache)) or force:
            if profile:
                msg="Reading events for Signature cache"
                profiler=Profiler(msg)
            sqlDB=EventStorage.getSqlDB()
            events=sqlDB.query("""select * from event""")
            if profiler:
                profiler.time()
            if profile:
                msg=f"Storing Signature cache for {len(events)} events"
                profiler=Profiler(msg)
            signature=SQLDB(signatureCache)
            entityInfo=signature.createTable(events, "event")
            signature.store(events, entityInfo)          
            if profiler:
                profiler.time()
            ddls = [
                "DROP INDEX if EXISTS eventsByCountry",
                "CREATE INDEX eventsByCounty ON event(country)"]
            for ddl in ddls:
                signature.execute(ddl)
      
        else:
            signature=SQLDB(signatureCache)
        return signature
    
    @classmethod
    def createLookup(cls,column:str,tables:list):
        '''
        create a lookup for a column for the given list of tables
        
        Args:
            column(str): the column to create the lookup for
            tables(str): the names of the tables to take into account
        '''
        sqlDB=EventStorage.getSqlDB()
        idColumn=f"{column}Wikidataid"
        total=0
        lookup={}
        for table in tables:
            totalQuery=f"SELECT count(*) AS total from event_{table}"
            totalRows=sqlDB.query(totalQuery)
            tableTotal=int(totalRows[0]["total"])
            total+=tableTotal
            sqlQuery=f"""SELECT COUNT(*) AS count,{column},{idColumn}
FROM event_{table}
GROUP BY {column},{idColumn}
ORDER BY 1 DESC"""
            lookupRows=sqlDB.query(sqlQuery)
            for lookupRow in lookupRows:
                entry=lookupRow[column]
                value=lookupRow["count"]
                qid=lookupRow[idColumn]
                if qid is not None:
                    if entry in lookup:
                        d=lookup[entry]
                        d["count"]+=value
                        d[table]=value
                    else:
                        lookup[entry]={"name":entry,"count":value,"qid":qid,f"{table}":value}   
            for entry in lookup:
                d=lookup[entry]
                d["frequency"]=round(d["count"]*100/total,3)
        return lookup 
    

class Event(JSONAble):
    '''
    base class for Event entities
    '''
    def __init__(self):
        '''
        Constructor
        '''
        super().__init__()

    def __str__(self):
        '''
        return my string representation
        
        Return:
            str: the string representation
        '''
        text=self.__class__.__name__
        attrs=["pageTitle","acronym","eventId","title","year","source","url"]
        delim=":"
        for attr in attrs:
            if hasattr(self, attr):
                value=getattr(self,attr)
                text+=f"{delim}{value}"
                delim=":" 
        return text
    
    def getLookupAcronym(self):
        ''' 
            get the lookup acronym of this event e.g. add year information 
            
        Return:
            str: the acronym to be used for lookup operations
        '''
        if hasattr(self,'acronym') and self.acronym is not None:
            self.lookupAcronym=self.acronym
        else:
            if hasattr(self,'event'):
                self.lookupAcronym=self.event
        if hasattr(self,'lookupAcronym'):
            if self.lookupAcronym is not None:
                try:
                    if hasattr(self, 'year') and self.year is not None and not re.search(r'[0-9]{4}',self.lookupAcronym):
                        self.lookupAcronym="%s %s" % (self.lookupAcronym,str(self.year))
                except TypeError as te:
                    print ('Warning getLookupAcronym failed for year: %s and lookupAcronym %s' % (self.year,self.lookupAcronym))   
    

    def getRecord(self):
        '''
        get my dict elements that are defined in getSamples
        
        Return:
            dict: fields of my __dict__ which are defined in getSamples
        '''
        fields = None
        if hasattr(self, 'getSamples') and callable(getattr(self, 'getSamples')):
            fields = LOD.getFields(self.getSamples())
        record = {}
        recordDict= self.__dict__
        for field in fields:
            if field in recordDict:
                record[field] = recordDict[field]
        return record
    
    def mapFromDict(self,d:dict,maptuples):
        '''
        set my attributes from the given dict mapping with the given
        mapping (key->attr) tuples
        
        Args:
            d(dict): the dictionary to map
            maptuples(list): the list of tuples for mapping
        '''
        for key,attr in maptuples:
            if key in d:
                setattr(self,attr,d[key])

    def asWikiMarkup(self,series:str,templateParamLookup:dict)->str:
        '''
        Args:
            series(str): the name of the series
            templateParamLookup(dict): the mapping of python attributes to Mediawiki template parameters to be used
        
        Return:
            str: my WikiMarkup
        '''
        nameValues={}
        delim=""
        for wikiName,attrName in templateParamLookup.items():
            if hasattr(self, attrName):
                value=getattr(self,attrName)
                nameValues[wikiName]=value
                
        markup=""
        nameValues["Series"]=series.upper()
        dblpConferenceId=re.sub(r"^https:\/\/dblp.org\/db\/conf\/","",self.url)
        dblpConferenceId=dblpConferenceId.replace(".html","")
        nameValues["DblpConferenceId"]=dblpConferenceId
        for name,value in nameValues.items():
            markup=f"{markup}{delim}|{name}={value}"
            delim="\n"
        markup=f"""{{{{Event
{markup}
}}}}"""
#|Type=Symposium

#|Submission deadline=2019/09/03
#|Homepage=http://ieeevr.org/2020/
#|City=Atlanta
#|Country=USA
#}}
        return markup
    
class EventSeries(JSONAble):
    '''
    base class for Event Series entities
    '''
    def __init__(self):
        '''
        Constructor
        '''
        super().__init__()
        
    def __str__(self):
        '''
        return my
        '''
        text=self.__class__.__name__
        attrs=["pageTitle","acronym","eventSeriesId","title","source","url"]
        delim=":"
        for attr in attrs:
            if hasattr(self, attr):
                value=getattr(self,attr)
                text+=f"{delim}{value}"
                delim=":" 
        return text
    
    def asWikiMarkup(self)->str:
        '''
        convert me to wikimarkup
        
        see https://github.com/WolfgangFahl/ConferenceCorpus/issues/10
        '''
        #dblpPid=self.DBLP_pid
        #if dblpPid:
        #    dblpPid=dblpPid.replace("conf/","")
        # |WikiDataId=
        #|Title={self.title}
        #|Homepage={self.homepage}
        markup=f"""{{{{Event series
|Acronym={self.acronym}
|DblpSeries={self.eventSeriesId}
}}}}"""
        #
        return markup


class EventBaseManager(EntityManager):
    '''
    common entity Manager for ConferenceCorpus
    '''
    
    def __init__(self,name,entityName,entityPluralName:str,listName:str=None,clazz=None,sourceConfig:EventDataSourceConfig=None,primaryKey:str=None,config=None,handleInvalidListTypes=False,filterInvalidListTypes=False,debug=False,profile=True):
        '''
        Constructor
        
        Args:
            name(string): name of this eventManager
            entityName(string): entityType to be managed e.g. Country
            entityPluralName(string): plural of the the entityType e.g. Countries
            config(StorageConfig): the configuration to be used if None a default configuration will be used
            handleInvalidListTypes(bool): True if invalidListTypes should be converted or filtered
            filterInvalidListTypes(bool): True if invalidListTypes should be deleted
            debug(boolean): override debug setting when default of config is used via config=None
            profile(boolean): True if profiling/timing information should be shown for long-running operations
        '''
        self.profile=profile
        if config is None:
            config=EventStorage.getStorageConfig(debug=debug)
            self.profile=config.profile
        if sourceConfig is not None:
            tableName=sourceConfig.getTableName(entityName)
        else:
            tableName=entityName
        super().__init__(name, entityName, entityPluralName, listName, clazz, tableName, primaryKey, config, handleInvalidListTypes, filterInvalidListTypes, listSeparator='⇹',debug=debug)
   
        
    def configure(self):
        '''
        configure me - abstract method that needs to be overridden
        '''    
        raise Exception(f"specialized configure for {self.name} needs to be implemented")
 
    def setAllAttr(self,listOfDicts,attr,value):
        '''
        set all attribute values of the given attr in the given list of Dict to the given value 
        '''
        for record in listOfDicts:
            record[attr]=value
            
    def rateAll(self,ratingManager:RatingManager):
        '''
        rate all events and series based on the given rating Manager
        '''
        for entity in self.getList():
            if hasattr(entity,"rate") and callable(entity.rate):
                if isinstance(entity,Event):
                    rating=EventRating(entity)
                elif isinstance(entity,EventSeries):
                    rating=EventSeriesRating(entity)
                else:
                    raise Exception(f"rateAll for unknown entity type {type(entity).__name__}")
                entity.rate(rating)
                ratingManager.ratings.append(rating)
            
    def fromCsv(self, csvString, separator:str= ',', overwriteEvents:bool = True, updateEntitiesCallback:Callable =None):
        """

        Args:
            csvString: csvString having all the csv content
            separator: the separator of the csv
            append: to append to the self object.
            updateEntitiesCallback:

        Returns: Nothing. The self object is upadated

        """
        fields= None
        # limit csv fields to the fields defined in the samples
        if hasattr(self.clazz, 'getSamples') and callable(getattr(self.clazz, 'getSamples')):
            fields = LOD.getFields(self.clazz.getSamples())
        eventRecords= CSV.fromCSV(csvString=csvString,fields=None,delimiter=separator)
        self.updateFromLod(eventRecords, overwriteEvents=overwriteEvents, updateEntitiesCallback=updateEntitiesCallback)

    def updateFromLod(self, lod:list, overwriteEvents:bool = True, updateEntitiesCallback:Callable=None, restrictToSamples:bool=True):
        """
        Updates the entities from the given LoD. If a entity does not already exist a new one will be added.
        Args:
            lod: data to update the entities
            overwriteEvents: If False only missing values are added
            updateEntitiesCallback: Callback function that is called on an updated entity
            restrictToSamples(bool): If True only properties that are names in the samples are set.

        Returns:

        """
        originalEventsLookup = self.getLookup(attrName=self.primaryKey)[0]
        for eventRecord in lod:
            if self.primaryKey in eventRecord:
                eventRecordPrimaryKey = eventRecord.get(self.primaryKey)
                if eventRecordPrimaryKey in originalEventsLookup:
                    originalEvent = originalEventsLookup[eventRecordPrimaryKey]
                    if hasattr(originalEvent, self.primaryKey):
                        sampleProperties = []
                        if hasattr(originalEvent, 'getSamples') and callable(originalEvent.getSamples):
                            sampleProperties = LOD.getFields(originalEvent.getSamples())
                        for key, value in eventRecord.items():
                            if hasattr(originalEvent, key):
                                setattr(originalEvent, key, value)
                            else:
                                if restrictToSamples or key in sampleProperties:
                                    setattr(originalEvent, key, value)
                                else:
                                    pass
                        if updateEntitiesCallback is not None and callable(updateEntitiesCallback):
                            updateEntitiesCallback(originalEvent, overwrite=overwriteEvents)
                else:
                    self.fromLoD(lod=[eventRecord], append=True, debug=self.debug)
                    # new entity was addded → update lookup
                    originalEventsLookup = self.getLookup(attrName=self.primaryKey)[0]
                    originalEvent = originalEventsLookup[eventRecordPrimaryKey]
                    if updateEntitiesCallback is not None and callable(updateEntitiesCallback):
                        updateEntitiesCallback(originalEvent, overwrite=overwriteEvents)

    def fromCache(self,force:bool=False,getListOfDicts=None,append=False,sampleRecordCount=-1):
        '''
        overwritten version of fromCache that calls postProcessEntityList
        '''
        needsUpdate=not self.isCached() or force
        super().fromCache(force, getListOfDicts, append, sampleRecordCount)
        if needsUpdate:
            # TODO
            # this is inefficient and uses 2x the memory 
            # try postProcessing on lod instead
            self.postProcessEntityList(debug=self.debug)
            self.store()
            
    def postProcessEntityList(self,debug:bool=False):
        '''
        postProcess my entities
        '''
        # override this method
        pass
        
    def asCsv(self, separator:str=',', selectorCallback:Callable=None):
        """
        Converts the events to csv format
        Args:
            separator(str): character separating the row values
            selectorCallback: callback functions returning events to be converted to csv. If None all events are converted.

        Returns:
            csv string of events
        """
        events=self.getList()
        if selectorCallback is not None and callable(selectorCallback):
            events=selectorCallback()
            if events and type(events) != list:
                events=[events]
        fields=None
        # limit csv fields to the fields defined in the samples
        if hasattr(self.clazz, 'getSamples') and callable(getattr(self.clazz, 'getSamples')):
            fields=LOD.getFields(self.clazz.getSamples())
        if events:
            csvString=CSV.toCSV(events, includeFields=fields, delimiter=separator)
            return csvString
        return None
    
    def postProcessLodRecords(self,listOfDicts:list,**kwArgs):
        '''
        post process the given list of Dicts with raw Events
        
        Args: 
            listOfDicts(list): the list of raw Events to fix
        '''
        if hasattr(self.clazz,"postProcessLodRecord") and callable(self.clazz.postProcessLodRecord): 
            for rawEvent in listOfDicts:
                self.clazz.postProcessLodRecord(rawEvent,**kwArgs)    
                
    def getLoDfromEndpoint(self)->list:
        '''
        get my content from my endpoint
        
        Returns:
            list: the list of dicts derived from the given SPARQL query
        '''
        sparql=SPARQL(self.endpoint)
        query=self.getSparqlQuery()
        try:
            profiler=Profiler(f"SPARQL query to {self.endpoint}",profile=False)
            listOfDicts=sparql.queryAsListOfDicts(query)
        except Exception as ex:
            # handle any Exception - e.g. there might be a syntax error in the query or the
            # endpoint might not be able to handle it - the endpoint might not be available
            # or there might be a timeout
            msg=f"SPARQL query failed\nquery:\n{query}"
            profiler.profile=True
            profiler.time()
            print(msg,file=sys.stderr)
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            exmessage = template.format(type(ex).__name__, ex.args)
            print(exmessage,file=sys.stderr)
            raise(ex)
        self.postProcessLodRecords(listOfDicts)
        self.setAllAttr(listOfDicts,"source",self.source)
        return listOfDicts

    def getEventByKey(self, keyToSearch, keytype='pageTitle'):
        for event in self.getList():
            if hasattr(event, keytype):
                if getattr(event, keytype) == keyToSearch:
                    return event
            else:
                raise ValueError("Invalid keytype given")


class EventSeriesManager(EventBaseManager):
    '''
    Event series list
    '''
    def __init__(self,name:str,sourceConfig:EventDataSourceConfig=None,clazz=None,primaryKey:str=None,config:StorageConfig=None,debug=False):
        '''
        constructor 
        '''
        super().__init__(name=name,entityName="EventSeries",entityPluralName="EventSeries",primaryKey=primaryKey,listName="series",clazz=clazz,sourceConfig=sourceConfig,handleInvalidListTypes=True,config=config,debug=debug) 
            
class EventManager(EventBaseManager):
    '''
    Event entity list
    '''
    
    def __init__(self,name:str,sourceConfig:EventDataSourceConfig=None,clazz=None,primaryKey:str=None,config:StorageConfig=None,debug=False):
        '''
        constructor 
        '''
        super(EventManager, self).__init__(name=name,entityName="Event",entityPluralName="Events",primaryKey=primaryKey,listName="events",clazz=clazz,sourceConfig=sourceConfig,config=config,handleInvalidListTypes=True,debug=debug,profile=config.profile if config else False)
        
 
    def linkSeriesAndEvent(self, eventSeriesManager:EventSeriesManager, seriesKey:str="series"):
        '''
        link Series and Event using the given foreignKey

        Args:
            seriesKey(str): the key to be use for lookup
            eventSeriesManager(EventSeriesManager):
        '''
        # get foreign key hashtable
        self.seriesLookup = LOD.getLookup(self.getList(), seriesKey, withDuplicates=True)
        # get "primary" key hashtable
        self.seriesAcronymLookup = LOD.getLookup(eventSeriesManager.getList(), "acronym", withDuplicates=True)

        for seriesAcronym in self.seriesLookup.keys():
            if seriesAcronym in self.seriesAcronymLookup:
                seriesEvents = self.seriesLookup[seriesAcronym]
                if hasattr(self, 'verbose') and self.verbose:
                    print(f"{seriesAcronym}:{len(seriesEvents):4d}")
            else:
                if self.debug:
                    print(f"Event Series Acronym {seriesAcronym} lookup failed")
        if self.debug:
            print("%d events/%d eventSeries -> %d linked" % (
            len(self.getList()), len(eventSeriesManager.getList()), len(self.seriesLookup)))
            
    def getEventsInSeries(self,seriesAcronym):
        """
        Return all the events in a given series.
        """
        if seriesAcronym in self.seriesLookup:
            seriesEvents = self.seriesLookup[seriesAcronym]
            if self.debug:
                print(f"{seriesAcronym}:{len(seriesEvents):4d}")
        else:
            if self.debug:
                print(f"Event Series Acronym {seriesAcronym} lookup failed - Series not known")
            return None
        return seriesEvents
    
    @staticmethod
    def asWikiSon(eventDicts):  
        wikison=""
        for eventDict in eventDicts:
            wikison+=EventManager.eventDictToWikiSon(eventDict)
        return wikison
    
    @staticmethod
    def eventDictToWikiSon(eventDict):
        wikison="{{Event\n"
        for key,value in eventDict.items():
            if key not in ['foundBy','source','creation_date','modification_date']:
                if value is not None:
                    wikison+="|%s=%s\n" % (key,value)
        wikison+="}}\n"  
        return wikison




'''
Created on 2022-02-16

@author: wf
'''

class Version(object):
    '''
    Version handling for ConferenceCorpus
    '''
    version='0.0.35'
    date = '2020-09-10'
    updated = '2022-08-24'

        


'''
Created on 2022-03-04

@author: wf
'''
from xml.etree.ElementTree import Element
from xml.etree import cElementTree as ElementTree
from xml.etree.ElementTree import ParseError
#from lxml.etree.ElementTree import ElementTree
from typing import Iterator
import sys
from corpus.utils.progress import Progress

class XmlEntity(object):
    '''
    an entity based on an XML object
    '''
    debug=False
    encoding="utf-8"
    
    def __init__(self,element:Element,xmlPropertyMap:dict,namespaces:dict):
        ''' 
        constructor
        
        Args:
            element: the element to construct me from
        '''
        self._props=[]
        if XmlEntity.debug:
            xml=ElementTree.tostring(element).decode(XmlEntity.encoding)
            # debug e.g. with http://xpather.com/
            print(xml)
            setattr(self,"rawxml",xml)
        for prop, xpath in xmlPropertyMap.items():
            attr=None
            if "@"in prop:
                propparts=prop.split("@")
                prop=propparts[0]
                attr=propparts[1]
            valueElements = element.findall(xpath,namespaces)
            if valueElements is not None and len(valueElements)>0:
                # single or multi?
                if len(valueElements)==1:
                    value=self.getValue(valueElements[0],attr) 
                else:
                    value=[]
                    for valueElement in valueElements:
                        value.append(self.getValue(valueElement,attr))
                setattr(self,prop,value)
                self._props.append(prop)
        pass        
    
    def getValue(self,element,attr:str=None):
        '''
        get the value from the element
        
        Args:
            element: the XML Node
            attr: the attribute to get the value for - if None get the text
        '''
        if attr is None:
            value=element.text
        else:
            if attr in element.attrib:
                value=element.attrib[attr]
            else:
                value=None
        return value
    
    def __str__(self):
        text=""
        for prop in self._props:
            text=f"{text}{prop}:{getattr(self,prop)}\n"
        return text
    
    def asDict(self):
        '''
        return me as a dictionary
        '''
        d={}
        for prop in self._props:
            d[prop]=getattr(self,prop)
        return d
    
class XMLEntityParser():
    '''
    a parser for XML Entities
    '''
    def __init__(self,filePath:str,recordsTag:str,progress:Progress=None):
        '''
        Constructor
        
        Args:
            filePath(str): the path to the xml file to parse
            recordsTag(str): the name of the tag to parse
            progressSteps(int): how often to show the progress of the parser
        '''
        self.filePath=filePath
        self.recordsTag=recordsTag
        self.progress=progress
        
    def readXmlFile(self) -> Iterator[Element]:
        '''
           Reads an XML file element by element and returns an iterator XML elements
           see https://github.com/sopherapps/xml_stream/blob/master/xml_stream/__init__.py 
        '''
        with open(self.filePath, 'rb') as xml_file:
            context = ElementTree.iterparse(xml_file, events=('start', 'end',))
            context = iter(context)
            root = None
    
            for event, element in context:
                if root is None:
                    root = element
    
                if event == 'end' and element.tag == self.recordsTag:
                    yield element
                    # clear the root element to leave it empty and use less memory
                    if root != element:
                        root.clear()
        
    def parse(self,xmlPropertyMap:dict,namespaces:dict)-> Iterator[XmlEntity]:
        '''
        parse my file
        
        Args:
            xmlPropertyMap(dict): attribute/xpath expression dict
            namespaces(dict): namespace / namespace path dict 
        '''
        try:
            for element in self.readXmlFile():
                yield XmlEntity(element,xmlPropertyMap,namespaces)
                if self.progress is not None:
                    self.progress.next()
        except ParseError as parseError:
            print (f"parse error in {self.filePath}:{parseError}", file=sys.stderr)
            pass


'''
Created on 2021-08-21

@author: wf
'''
import os
import urllib
import gzip
import shutil
import time

class Download:
    '''
    Utility functions for downloading data
    '''
    
    @staticmethod
    def getURLContent(url:str):
        with urllib.request.urlopen(url) as urlResponse:
            content = urlResponse.read().decode()
            return content

    @staticmethod
    def getFileContent(path:str):
        with open(path, "r") as file:
            content = file.read()
            return content

    @staticmethod
    def needsDownload(filePath:str,force:bool=False)->bool:
        '''
        check if a download of the given filePath is necessary that is the file
        does not exist has a size of zero or the download should be forced
        
        Args:
            filePath(str): the path of the file to be checked
            force(bool): True if the result should be forced to True
            
        Return:
            bool: True if  a download for this file needed
        '''
        if not os.path.isfile(filePath):
            result=True
        else:
            stats=os.stat(filePath)
            size=stats.st_size
            result=force or size==0
        return result

    @staticmethod
    def downloadBackupFile(url:str, fileName:str, targetDirectory:str, force:bool=False,profile:bool=True):
        '''
        Downloads from the given url the zip-file and extracts the file corresponding to the given fileName.

        Args:
            url: url linking to a downloadable gzip file
            fileName: Name of the file that should be extracted from gzip file
            targetDirectory(str): download the file to this directory
            force (bool): True if the download should be forced
            profile(bool): if True show profiling information

        Returns:
            Name of the extracted file with path to the backup directory
        '''
        extractTo = f"{targetDirectory}/{fileName}"
        # we might want to check whether a new version is available
        if Download.needsDownload(extractTo,force=force):
            if not os.path.isdir(targetDirectory):
                os.makedirs(targetDirectory)
            zipped = f"{extractTo}.gz"
            msg=f"Downloading {zipped} from {url} ... this might take a few seconds ..."
            profiler=Profiler(msg=msg,profile=profile)
            urllib.request.urlretrieve(url, zipped)
            profiler.time(extraMsg=f" unzipping {extractTo} from {zipped}")
            with gzip.open(zipped, 'rb') as gzipped:
                with open(extractTo, 'wb') as unzipped:
                    shutil.copyfileobj(gzipped, unzipped)
            if not os.path.isfile(extractTo):
                raise (f"could not extract {fileName} from {zipped}")
        return extractTo
    
class Profiler:
    '''
    simple profiler
    '''
    def __init__(self,msg:str=None,profile=True):
        '''
        construct me with the given msg and profile active flag
        
        Args:
            msg(str): the message to show if profiling is active
            profile(bool): True if messages should be shown
        '''
        if msg is not None:
            self.msg=msg
        else:
            self.msg=""
        self.profile=profile
        self.starttime=time.time()
        if profile:
            print(f"Starting {msg} ...")
    
    def time(self,extraMsg=""):
        '''
        time the action and print if profile is active
        '''
        elapsed=time.time()-self.starttime
        if self.profile:
            print(f"{self.msg}{extraMsg} took {elapsed:5.1f} s")
        return elapsed
        


import re
from enum import Enum


class IsbnType(Enum):
    ISBN_10 = "ISBN-10"
    ISBN_13 = "ISBN-13"
    NONE = None


class ISBN:
    """
    Provides Functions to detect the ISBN version and convert between them if possible
    """

    # https://www.wikidata.org/wiki/Property:P957
    ISBN_10_REGEX_PATTERN = "([0-57]-(\d-\d{7}|\d\d-\d{6}|\d{3}-\d{5}|\d{4}-\d{4}|\d{5}-\d{3}|\d{6}-\d\d|\d{7}-\d)|(" \
                            "65|8\d|9[0-4])-(\d-\d{6}|\d\d-\d{5}|\d{3}-\d{4}|\d{4}-\d{3}|\d{5}-\d\d|\d{6}-\d)|(6[" \
                            "0-4]|9[5-8])\d-(\d-\d{5}|\d\d-\d{4}|\d{3}-\d{3}|\d{4}-\d\d|\d{5}-\d)|99[0-8]\d-(\d-\d{" \
                            "4}|\d\d-\d{3}|\d{3}-\d\d|\d{4}-\d)|999\d\d-(\d-\d{3}|\d\d-\d\d|\d{3}-\d))-[\dX]"

    # https://www.wikidata.org/wiki/Property:P212
    ISBN_13_REGEX_PATTERN = "97(8-([0-57]-(\d-\d{7}|\d\d-\d{6}|\d{3}-\d{5}|\d{4}-\d{4}|\d{5}-\d{3}|\d{6}-\d\d|\d{" \
                            "7}-\d)|(65|8\d|9[0-4])-(\d-\d{6}|\d\d-\d{5}|\d{3}-\d{4}|\d{4}-\d{3}|\d{5}-\d\d|\d{" \
                            "6}-\d)|(6[0-4]|9[5-8])\d-(\d-\d{5}|\d\d-\d{4}|\d{3}-\d{3}|\d{4}-\d\d|\d{5}-\d)|99[" \
                            "0-8]\d-(\d-\d{4}|\d\d-\d{3}|\d{3}-\d\d|\d{4}-\d)|999\d\d-(\d-\d{3}|\d\d-\d\d|\d{" \
                            "3}-\d))|9-(8-(\d-\d{7}|\d\d-\d{6}|\d{3}-\d{5}|\d{4}-\d{4}|\d{5}-\d{3}|\d{6}-\d\d|\d{" \
                            "7}-\d)|(1[0-2]|6\d)-(\d-\d{6}|\d\d-\d{5}|\d{3}-\d{4}|\d{4}-\d{3}|\d{5}-\d\d|\d{" \
                            "6}-\d)))-\d|"

    @classmethod
    def getType(cls, value:str) -> IsbnType:
        if re.fullmatch(cls.ISBN_10_REGEX_PATTERN, value):
            return IsbnType.ISBN_10
        elif re.fullmatch(cls.ISBN_13_REGEX_PATTERN, value):
            return IsbnType.ISBN_13
        else:
            return IsbnType.NONE

    @classmethod
    def convertToIsbn13(cls, value) -> str:
        """
        If given value is a valid ISBN-10 the value is converted to ISBN-13.
        If the given value is already a valid ISBN-13 the value is returned.
        Otherwise, None is returned

        see https://en.wikipedia.org/wiki/ISBN#ISBN-10_to_ISBN-13_conversion
        Args:
            value: isbn value

        Returns:
            str ISBN-13 value or None if the value can not be converted to a ISBN-13 value
        """
        isbn_type = cls.getType(value)
        if isbn_type is IsbnType.ISBN_10:
            value = f"978-{value}"
        elif isbn_type is IsbnType.NONE:
            value = None
        elif isbn_type is IsbnType.ISBN_13:
            pass
        return value



'''
Created on 2021-07-31

@author: wf
'''

from corpus.event import EventStorage,EventSeriesManager, EventSeries, Event, EventManager
from lodstorage.storageconfig import StorageConfig
import corpus.datasources.wikicfpscrape
from corpus.eventcorpus import EventDataSource, EventDataSourceConfig
from corpus.quality.rating import Rating, RatingType
from datetime import datetime
from corpus.datasources import wikicfpscrape
from ptp.ordinal import Ordinal

class WikiCfp(EventDataSource):
    '''
    scientific event from http://www.wikicfp.com
    '''
    sourceConfig = EventDataSourceConfig(lookupId="wikicfp", name="WikiCFP", url='http://www.wikicfp.com', title='WikiCFP', tableSuffix="wikicfp",locationAttribute="locality")
    
    def __init__(self,debug=False):
        '''
        constructor
        '''
        super().__init__(WikiCfpEventManager(), WikiCfpEventSeriesManager(), WikiCfp.sourceConfig)
        self.debug=debug
        config=EventStorage.getStorageConfig(mode='json')
        jsonEventCache=WikiCfpEventManager(config=config)
        jsonEventSeriesCache=WikiCfpEventSeriesManager(config=config)
        self.wikiCfpScrape=corpus.datasources.wikicfpscrape.WikiCfpScrape(jsonEventCache,jsonEventSeriesCache)
    
class WikiCfpEvent(Event):
    '''
    event derived from WikiCFP
    '''
    
    @classmethod
    def getSamples(cls):
        samples = [
        {
            "Notification_Due": datetime.fromisoformat("2008-05-29"),
            "Submission_Deadline": datetime.fromisoformat("2008-04-15"),
            "acronym": "SIGMAP 2008",
            "city": "Porto",
            "country": "Portugal",
            "deleted": False,
            "endDate": datetime.fromisoformat("2008-07-29"),
            "eventType": "Conference",
            "homepage": "http://www.sigmap.org/SIGMAP2008/CallforPapers.html",
            "locality": "Poato, Portugal",
            "lookupAcronym": "SIGMAP 2008",
            "source": "wikicfp",
            "startDate": datetime.fromisoformat("2008-07-26"),
            "title": "SIGMAP  2008 : International Conference on Signal Processing and Multimedia Applications",
            "url": "http://www.wikicfp.com/cfp/servlet/event.showcfp?eventid=977",
            "wikiCFPId": 977,
            "year": 2008
        }]
        return samples
    
    @staticmethod
    def postProcessLodRecord(rawEvent):
        '''
        postProcess the given rawEvent
        
        Args:
            rawEvent(dict): the record to post process
        '''
        rawEvent["source"]="wikicfp"
        if not "url" in rawEvent:
            crawlType=wikicfpscrape.CrawlType.EVENT
            eventId=rawEvent["eventId"]
            rawEvent["url"]=f"{crawlType.urlPrefix}{eventId}"
        Ordinal.addParsedOrdinal(rawEvent)
        
    
    def rate(self,rating:Rating):
        '''
        rate me
        
        Args:
            rating(Rating): the rating to modify
        '''
        rating.set(0, RatingType.ok, "")
        
class WikiCfpEventSeries(EventSeries):
    '''
    event series derived from WikiCFP
    '''
    
    @classmethod
    def getSamples(cls):
        samples = [
            {
                "dblpSeriesId": "conf/aaai",
                "seriesId": "3",
                "title": "AAAI: National Conference on Artificial Intelligence 2022 2021 2020 ...",
                "wikiCfpId": 3
            }
        ]
        return samples
    
    @staticmethod
    def postProcessLodRecord(rawEvent):
        '''
        postProcess the given rawEvent
        
        Args:
            rawEvent(dict): the record to post process
        '''
        rawEvent["source"]="wikicfp"
        rawEvent["eventSeriesId"]=rawEvent["seriesId"]
        if not "url" in rawEvent:
            crawlType=wikicfpscrape.CrawlType.SERIES
            seriesId=rawEvent["seriesId"]
            rawEvent["url"]=f"{crawlType.urlPrefix}{seriesId}"
        if "title" in rawEvent:
            title=rawEvent["title"]
            if title:
                if ":" in title:
                    rawEvent["acronym"]=title.split(":")[0]
                suffix="2022 2021 2020 ..."
                if title.endswith(suffix):
                    title=title.replace(suffix,"")
            rawEvent["title"]=title
    
    
class WikiCfpEventManager(EventManager):
    '''
    manage WikiCFP derived scientific events
    '''

    def __init__(self, config:StorageConfig=None):
        '''
        Constructor
        '''
        super().__init__(name="WikiCfpEvents", sourceConfig=WikiCfp.sourceConfig,clazz=WikiCfpEvent, config=config)
        
    def configure(self):
        '''
        configure me
        '''
        # no need - getListOfDicts has a specialized implementation
        pass
        
    def getListOfDicts(self):
        '''
        get my list of dicts
        '''
        lod = []
        if  hasattr(self, "dataSource"):
            jsonEm=self.dataSource.wikiCfpScrape.cacheToJsonManager(corpus.datasources.wikicfpscrape.CrawlType.EVENT) 
            for event in jsonEm.events:
                lod.append(event.__dict__)
            self.postProcessLodRecords(lod)
        return lod    

 
class WikiCfpEventSeriesManager(EventSeriesManager):
    '''
    manage WikiCFP derived scientific conference series
    '''

    def __init__(self, config:StorageConfig=None):
        '''
        Constructor
        '''
        super(WikiCfpEventSeriesManager, self).__init__(name="WikiCfpEventSeries", sourceConfig=WikiCfp.sourceConfig, clazz=WikiCfpEventSeries, config=config)
        
    def configure(self):
        '''
        configure me
        '''
        # no need - getListOfDicts has a specialized implementation
        
    def getListOfDicts(self):
        '''
        get my list of dicts
        '''
        lod = []
        if  hasattr(self, "dataSource"):
            jsonEm=self.dataSource.wikiCfpScrape.cacheToJsonManager(corpus.datasources.wikicfpscrape.CrawlType.SERIES)
            for series in jsonEm.series:
                lod.append(series.__dict__)
            self.postProcessLodRecords(lod)
        return lod    



'''
Created on 2021-01-25

@author: wf
'''
from pathlib import Path
from io import BytesIO
import urllib.request
from gzip import GzipFile
from lxml import etree
from collections import Counter
from xml.dom import minidom
from lodstorage.sql import SQLDB
from lodstorage.schema import Schema
from corpus.utils.progress import Progress
import os
import re
import time

class DblpXml(object):
    '''
    handler for https://dblp.uni-trier.de/xml/ dumps
    see https://github.com/IsaacChanghau/DBLPParser/blob/master/src/dblp_parser.py
    '''

    def __init__(self,xmlname:str="dblp.xml",dtd_validation:bool=False,xmlpath:str=None,gzurl:str="https://dblp.uni-trier.de/xml/dblp.xml.gz",debug=False,verbose=True):
        '''
        Constructor
        
        Args:
            xmlname (str): name of the xml file
            dtd_validation (bool): True if dtd validation should be activated when parsing
            xmlpath(str): download path
            gzurl(str): url of the gzipped original file
            debug(bool): if True show debugging information
            verbose(bool): if True show logging information
        '''
        self.debug=debug
        self.verbose=verbose
        if xmlpath is None:
            home = str(Path.home())
            xmlpath=f"{home}/.dblp" 
        self.gzurl=gzurl
        self.xmlname=xmlname
        self.xmlpath=xmlpath
        self.dtd_validation=dtd_validation
        self.reinit()
        
    def reinit(self):
        '''
        reinitialize my file names
        '''
        self.xmlfile="%s/%s" % (self.xmlpath,self.xmlname)
        self.dtdfile="%s/%s" % (self.xmlpath,self.xmlname.replace(".xml",".dtd"))
     
    def getSize(self)->int:
        '''
        get the size of my xmlFile
        
        Returns:
            int: the size 
        '''
        stats=os.stat(self.xmlfile)
        size=stats.st_size
        return size
    
    def getExpectedTotal(self)->int:
        '''
        get the expected Total of records
        '''
        return self.getSize()//380
    
    def warnFullSize(self):
        '''
        warn if we are using the full dataset
        '''
        print(f"Warning - using full {self.xmlfile} dataset ~{self.getExpectedTotal()/1000000:3.1f}m records!")  

    def isDownloaded(self,minsize:int=3000000000)->bool:
        '''
        check that the dblp file is downloaded
        
        Returns:
            bool: True if the dblpfile is fully downloaded and is bigger than the given minimum size
        '''
        result=os.path.isfile(self.xmlfile)
        if result:
            result=self.getSize()>=minsize
        return result
    
    def prettyXml(self,tree,indent='  '):
        '''
        get a pretty XML representation of the given etree
        '''
        xmlstr = minidom.parseString(etree.tostring(tree.getroot())).toprettyxml(indent=indent)
        return xmlstr
    
    def createSample(self,keyEntities=None,keyPrefix="conf/",entityLimit=1000,entities=None,progress:int=500000):
        '''
        create a sample with the given entityLimit
        
        Args:
            keyPrefix(str): the keyPrefix to filter for
        '''
        if entities is None:
            entities=['article','book','incollection','www']
        if keyEntities is None:
            keyEntities=['proceedings','inproceedings']
        allEntities=[]
        allEntities.extend(entities)
        allEntities.extend(keyEntities)
        root = etree.Element('dblp')
        counter=Counter()
        level=0
        showProgress=Progress(progress)
        for event, element in self.iterParser():
            showProgress.next()
            if event == 'start': 
                level += 1
                if level==2:
                    doadd=element.tag in entities
                    if element.tag in keyEntities:
                        if 'key' in element.attrib:
                            key=element.attrib['key']
                            if key.startswith(keyPrefix):
                                doadd=True
                    if (doadd and counter[element.tag]<entityLimit):
                        node=etree.fromstring(etree.tostring(element))
                        root.append(node)
                        counter[element.tag]+=1
                    else:
                        keys=counter.keys()
                        done=True
                        for entity in allEntities:
                            if not entity in keys:
                                done=False
                            else:
                                done=done and counter[entity]>=entityLimit
                        if done:
                            break
                        
                pass
            elif event == 'end':
                level -=1
            self.clear_element(element)
        sampleTree=etree.ElementTree(root) 
        return sampleTree
        
    def getXmlFile(self,reload=False):
        '''
        get the dblp xml file - will download the file if it doesn't exist
        
        Args:
            reload(bool): if True force download
            
        Returns:
            str: the xmlfile
        '''
        if not os.path.isfile(self.xmlfile) or reload:
            os.makedirs(self.xmlpath,exist_ok=True)
            if self.verbose:
                print(f"downloading {self.xmlfile} from {self.gzurl}")
            urlreq = urllib.request.urlopen(self.gzurl)
            z = GzipFile(fileobj=BytesIO(urlreq.read()), mode='rb')
            with open(self.xmlfile, 'wb') as outfile:
                outfile.write(z.read())
        if not os.path.isfile(self.dtdfile) or reload:
            dtdurl=self.gzurl.replace(".xml.gz",".dtd")
            urllib.request.urlretrieve (dtdurl, self.dtdfile)
        
        return self.xmlfile
    
    def iterParser(self):
        """
           Create a dblp data iterator of (event, element) pairs for processing
           Returns:
               etree.iterparse result
        """
        if not os.path.isfile(self.xmlfile):
            raise ("dblp xml file %s not downloaded yet - please call getXmlFile first")
        # with dtd validation
        if self.debug:
            print(f"starting parser for {self.xmlfile}"  )
        # https://lxml.de/api/lxml.etree.iterparse-class.html
        self.parser=etree.iterparse(source=self.xmlfile, events=('end', 'start' ), dtd_validation=self.dtd_validation, load_dtd=True, huge_tree=True) 
        return self.parser 
    
    def clear_element(self,element):
        """
        Free up memory for temporary element tree after processing the element
        
            Args:
                element(node): the etree element to clear together with it's parent
        """
        element.clear()
        while element.getprevious() is not None:
            del element.getparent()[0]
        
    def checkRow(self,kind:str,index,row:dict):
        '''
        check the row content
        
        Args:
            kind(str): e.g. proceedings/article
            index(int): the index of the row
            row(dict): the row to process
        '''        
        if kind=='proceedings':
            if 'title' in row:
                title=row['title']
                if not title:
                    print(f'empty title for {index}{row}')
            else:
                print(f'missing title for {index}{row}')
              
    def postProcess(self,_kind:str,_index,row:dict):
        '''
        postProcess the given row
        
        Args:
            _kind(str): e.g. proceedings/article
            _index(int): the index of the row
            row(dict): the row to process
        '''
        if 'key' in row:
            key=row['key']
            if key.startswith("conf/"):
                conf=re.sub(r"conf/(.*)/.*",r"\1",key)
                row['conf']=conf
        pass
    
    def getXmlSqlDB(self,reload=False,showProgress=False):
        '''
        get the SqlDB derived from the XML download 
        '''
        self.getXmlFile(reload=reload)
        return self.getSqlDB(postProcess=self.postProcess,showProgress=showProgress)
        
            
    def getSqlDB(self,limit=1000000000,sample=None,createSample=10000000,debug=False,recreate=False,postProcess=None,check_same_thread=False,showProgress:bool=False):
        '''
        get the SQL database or create it from the XML content
        
        Args:
            limit(int): maximum number of records
        '''
        dbname=f"{self.xmlpath}/dblp.sqlite"
        # estimate size
        if showProgress:
            expectedTotal=self.getExpectedTotal()
            progress=expectedTotal//86
        else:
            expectedTotal=None
            progress=None
        if sample is None:
            sample=5
        if (os.path.isfile(dbname)) and not recreate:
            sqlDB=SQLDB(dbname=dbname,debug=debug,errorDebug=True,check_same_thread=check_same_thread)
        else:
            if (os.path.isfile(dbname)) and recreate:
                os.remove(dbname)
            sqlDB=SQLDB(dbname=dbname,debug=debug,errorDebug=True,check_same_thread=check_same_thread)
            starttime=time.time()
            dictOfLod=self.asDictOfLod(limit,progressSteps=progress,expectedTotal=expectedTotal)
            elapsed=time.time()-starttime
            executeMany=True
            if showProgress:
                print(f"parsing done after {elapsed:5.1f} s ... storing ...")
            starttime=time.time()    
            fixNone=True    
            for i, (kind, lod) in enumerate(dictOfLod.items()):
                if postProcess is not None:
                    for j,row in enumerate(lod):
                        postProcess(kind,j,row)
            rows=0
            for i, (kind, lod) in enumerate(dictOfLod.items()):
                rows+=len(lod)
                if debug:
                    print ("#%4d %5d: %s" % (i+1,len(lod),kind))
                entityInfo=sqlDB.createTable(lod,kind,'key',sampleRecordCount=createSample,failIfTooFew=False)
                sqlDB.store(lod,entityInfo,executeMany=executeMany,fixNone=fixNone)
                for j,row in enumerate(lod):
                    if debug:
                        print ("  %4d: %s" % (j,row)) 
                    if j>sample:
                        break
            elapsed=time.time()-starttime        
            if showProgress:
                print (f"stored {rows} rows in {elapsed:5.1f} s {rows/elapsed:5.0f} rows/s" )
            tableList=sqlDB.getTableList()     
            viewDDL=Schema.getGeneralViewDDL(tableList, "record")
            if debug:
                print(viewDDL)
            sqlDB.execute(viewDDL)
        return sqlDB
            
    def asDictOfLod(self,limit:int=1000,delim:str=',',progressSteps:int=None,expectedTotal:int=None):
        '''
        get the dblp data as a dict of list of dicts - effectively separating the content
        into table structures
        
        Args:
            limit(int): maximum amount of records to process
            delim(str): the delimiter to use for splitting attributes with multiple values (e.g. author)
            progressSteps(int): if set the interval at which to print a progress dot 
            expectedTotal(int): the expected Total number 
        '''
        index=0
        progress=Progress(progressSteps,expectedTotal,msg="Parsing dblp xml dump",showMemory=True)
        level=0
        dictOfLod={}
        current={}
        levelCount=Counter()
        for event, elem in self.iterParser():
            if event == 'start': 
                level += 1
                levelCount[level]+=1
                if level==2:
                    kind=elem.tag
                    if not kind in dictOfLod:
                        dictOfLod[kind]=[]
                    lod=dictOfLod[kind]
                    # copy the attributes (if any)
                    if hasattr(elem, "attrib"):
                        current = {**current, **elem.attrib}
                elif level==3:
                    name=elem.tag
                    newvalue=elem.text
                    # is there already an entry for the given name
                    if name in current:
                        oldvalue=current[name]
                        newvalue=f"{oldvalue}{delim}{newvalue}"
                    # set the name/value pair
                    current[name]=newvalue    
                    if (kind=="proceedings") and (name=="title") and (elem.text is None):
                        print(f"{elem.sourceline:6}:{elem.tag} - None text")
                        pass
                elif level>=4:
                    # interesting things happen here ...
                    # sub/sup i and so on see dblp xml faq
                    #if elem.sourceline:
                    #    print(f"{elem.sourceline:6}:{elem.tag}")
                    pass
            elif event == 'end':
                if level==2:
                    lod.append(current)
                    progress.next()
                    kind=elem.tag
                    self.checkRow(kind,progress.count,current)
                    current={} 
                    if progress.count>=limit:
                        break
                level -= 1
                self.clear_element(elem)
                index+=1
        if self.debug:
            pass
        if progress is not None:
            progress.done()
        return dictOfLod



'''
Created on 2022-03-04

@author: wf
'''
import itertools
from wikibot.wikipush import WikiPush
from wikifile.wikiFile import WikiFile
from wikifile.wikiFileManager import WikiFileManager

from corpus.datasources.openresearch import OREvent
from corpus.version import Version
import json
import os
import sys
DEBUG = 1
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from corpus.xmlhandler.xmlparser import XMLEntityParser, XmlEntity
from typing import Iterator, Union
from corpus.utils.progress import Progress


class FTXParser(object):
    '''
    an specialized XML parser for FTX dump files
    '''
    
    def __init__(self,basepath:str):
        '''
        constructor
        
        Args:
            basepath(str): the root of the FTX dump directory
        '''
        self.basepath=basepath
        
    def ftxXmlFile(self,xmlFile:str)->str:
        '''
        Args: 
            xmlFile(str): the name of the FTX XML file 
            
        Returns:
            str: the full path
        
        '''
        path=f"{self.basepath}/{xmlFile}"
        return path
        
    def ftxXmlFiles(self)->list:
        '''
        get the relevant XML files list
        '''
        files = os.listdir(self.basepath)
        files = [f for f in files if f.endswith(".xml")]
        return files
    
    def parse(self,xmlFile,local:bool=True,progress:Progress=None)->Iterator[XmlEntity]:
        '''
          parse the xml data of  volume with the given collectionId
          
          Args:
            xmlFile(str): the file to process
            local(bool): True if the xmlFile name is a local file name
        '''    
        recordTag="{http://www.openarchives.org/OAI/2.0/}document"
        namespaces={'ns0':'http://www.openarchives.org/OAI/2.0/',
                    'ns2': 'http://purl.org/dc/terms/',
                    'dc':'http://purl.org/dc/elements/1.1/'
        }
        xmlPropertyMap= {
            "databaseDate": './ns0:systemInfo/ns0:databaseDate',
            "changeDate": './ns0:systemInfo/ns0:changeDate',
            "ftxCreationDate": './ns0:systemInfo/ns0:ftxCreationDate',
            # formalInfo
            "documentGenreCode": './ns0:formalInfo//ns0:documentGenreCode',
            "documentTypeCode": './ns0:formalInfo//ns0:documentTypeCode',
            # identifiers
            "documentId": './ns0:systemInfo/ns0:documentID',
            "ppn": './/ns0:identifiers/ns0:identifier[@type="ppn"]',
            "firstid": './/ns0:identifiers/ns0:identifier[@type="firstid"]',
            "isbn": './/ns0:identifiers/ns0:identifier[@type="isbn"]',
            "isbn13": './/ns0:identifiers/ns0:identifier[@type="isbn13"]',
            "ean": './/ns0:identifiers/ns0:identifier[@type="ean"]',
            "doi": './/ns0:identifiers/ns0:identifier[@type="doi"]',
            # from corporate creators
            # check for type not existing to get gnd ID
            # https://stackoverflow.com/a/13807791/1497139
            "corporateCreatorTypes@type": './/ns0:corporateCreator',
            "corporateCreatorNames": './/ns0:corporateCreator//ns0:name',
            "gndIds": './/ns0:corporateCreator/ns0:corporateIDs/ns0:corporateID[@type="gnd"]',
            "authorGndId": './/ns0:corporateCreator[@type="author"]/ns0:corporateIDs/ns0:corporateID[@type="gnd"]',
            "sponsorGndId": './/ns0:corporateCreator[@type="sponsor"]/ns0:corporateIDs/ns0:corporateID[@type="gnd"]',
            # bibliographic info
            "title": './ns0:bibliographicInfo/dc:title',
            "alternativeTitles": './ns0:bibliographicInfo/ns0:alternativeTitles/ns2:alternative',
            # conferenceInfo
            "event":  './/ns0:conferenceInfo/ns0:name',
            "location": './/ns0:conferenceInfo/ns0:places/ns0:place',
            "dates":  './/ns0:conferenceInfo/ns0:dates/dc:date',
            "description": './/ns0:conferenceInfo/dc:description',
            "pubyear": './ns0:bibliographicInfo/ns0:publicationInfo/ns2:issued',
            "pubplace": './ns0:bibliographicInfo/ns0:publicationInfo/ns0:publicationPlaces/ns0:publicationPlace',
            # "pubcountry": './ns0:bibliographicInfo/ns0:publicationInfo/ns0:publicationCountries/ns0:publicationCountry', # code Attribute
            "publisher": './ns0:bibliographicInfo/ns0:publicationInfo/dc:publisher',
            # journalInfo
            "journalTitle": './/ns0:journalInfo/dc:title',
            "journalVolumeNumber": './/ns0:journalInfo/ns0:volumeNumber',
            # classification Info
            "bk": "./ns0:classificationInfo/ns0:classifications/ns0:classification[@classificationName='bk']/ns0:code",
            "ddc": "./ns0:classificationInfo/ns0:classifications/ns0:classification[@classificationName='ddc']/ns0:code"
        }        
        if local:
            xmlPath=self.ftxXmlFile(xmlFile)
        else:
            xmlPath=xmlFile
        if os.path.exists(xmlPath):
            xmlParser=XMLEntityParser(xmlPath,recordTag)
            for xmlEntity in xmlParser.parse(xmlPropertyMap,namespaces):
                yield(xmlEntity)
                if progress is not None:
                    progress.next()


class TibkatCmdLine:
    '''
    TIB Kat command line 
    '''
    @property
    def debug(self):
        return getattr(self, "_debug", False)

    @debug.setter
    def debug(self, debug:bool):
        setattr(self, "_debug", debug)
    
    def main(self,args):
        '''
        command line access
        '''
        # should take json import and modify target wiki
        self.debug=args.debug
        wikiId = args.target
        # read from stdin
        if args.file is not None:
            path = os.path.abspath(args.file)
            if os.path.isfile(path):
                with open(path, mode="r") as f:
                    rawInput = f.read()
            else:
                print(f"Given file does not exits! ({path})")
                return
        else:
            rawInput = sys.stdin.read()
        data = json.loads(rawInput)
        records = data
        stats={}
        total = len(records)
        for i, record in enumerate(records):
            if self.debug:
                print(f"{i}/{total}:\t {record.get('title', '<Has no title>')}")
            pageTitle, ppnId = self.addPpnIdToWiki(wikiId, record, args.dryrun)
            if pageTitle is None:
                continue
            if pageTitle in stats:
                stats[pageTitle] += [ppnId]
            else:
                stats[pageTitle] = [ppnId]
        # print stats
        wikipush = WikiPush(fromWikiId=wikiId)
        wikiUrl = wikipush.fromWiki.wikiUser.url[:-1] + wikipush.fromWiki.wikiUser.scriptPath
        for pageTitle, pnnIds in stats.items():
            if len(pnnIds) == 1:
                print(f"{bcolors.OKGREEN}{pageTitle}({wikiUrl}/) found Id: {pnnIds}{bcolors.ENDC}")
            elif len(pnnIds) > 1:
                print(f"{bcolors.WARNING}{pageTitle}({wikiUrl}/{pageTitle.replace(' ', '_')}) found Ids: {pnnIds}{bcolors.ENDC}")
            else:
                print(f"{bcolors.OKCYAN}{pageTitle}({wikiUrl}/) no id in records {bcolors.ENDC}")

    def addPpnIdToWiki(self, wikiId:str, record:dict, isDryRun:bool=False) -> (str,str):
        """
        tries to find a matching event in the given wiki for the given event records to add the ppnId

        Args:
            wikiId: id of the wiki
            record: event record containing event acronym, ordinal and ppn

        Returns:
            (str, str) (name of the matching page, ppn that was added)
        """
        event = record.get("event", None)
        if event is None:
            return None, None
        eventParts = event.split(";")
        if len(eventParts) < 2:
            return None, None
        seriesAcronym = eventParts[0].strip()
        ordinal = eventParts[-1].strip()
        wikiFile = self.getMatchingEventFromWiki(wikiId, seriesAcronym, ordinal)
        if wikiFile is None:
            return None, None
        else:
            ppnId = record.get("ppn")
            if ppnId is None:
                return wikiFile.getPageTitle(), None
            ppnTemplateName = [r.get("templateParam") for r in OREvent.propertyLookupList if r.get("name") == "TibKatId"][0]
            wikiFile.updateTemplate(OREvent.templateName, args={ppnTemplateName:ppnId}, overwrite=True)
            if not isDryRun:
                wikiFile.pushToWiki("Added TibKatId")
            return wikiFile.getPageTitle(), ppnId

    def getMatchingEventFromWiki(self, wikiId:str, acronym:str, ordinal:int) -> Union[WikiFile, None]:
        """
        retrieves a matching event form the given wiki and returns the wikiFile
        Args:
            wikiId: id of the wiki
            acronym: acronym of the event series
            ordinal: ordinal of the event

        Returns:
            WikiFile if a matching event is found otherwise None
        """
        manager = WikiFileManager(sourceWikiId=wikiId, targetWikiId=wikiId)
        query = f"{{{{#ask: [[Event in series::{acronym}]][[Ordinal::{ordinal}]]}}}}"
        qres = list(manager.wikiPush.query(query))
        if qres:
            wikiFile = manager.getWikiFileFromWiki(pageTitle=qres[0])
            return wikiFile
        return None


class bcolors:
    """
    see https://stackoverflow.com/questions/287871/how-to-print-colored-text-to-the-terminal
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    
def main(argv=None): # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv=sys.argv[1:]
        
    program_name = os.path.basename(__file__)
    program_version = "v%s" % Version.version
    program_build_date = str(Version.updated)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = "script import TIBKat Data into openresearch compatible wiki"
    user_name="Wolfgang Fahl/Tim Holzheim"
    program_license = '''%s

  Created by %s on %s.
  Copyright 2022 contributors. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc,user_name, str(Version.date))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug", dest="debug",   action="store_true", help="set debug [default: %(default)s]")
        parser.add_argument("-t", "--target", dest="target", help="target wiki id")
        parser.add_argument("-f", "--file", dest="file", help="file with the data. If not defined read from stdin")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument('--dryrun', action="store_true", dest='dryrun', help="")
        args = parser.parse_args(argv)
        tibkatCmd=TibkatCmdLine()
        tibkatCmd.main(args)
     
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2     

if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())



'''
Created on 2021-07-28

@author: th
'''
from corpus.event import EventSeriesManager, EventSeries, Event, EventManager
from lodstorage.storageconfig import StorageConfig
from corpus.datasources.dblpxml import DblpXml
from corpus.eventcorpus import EventDataSource, EventDataSourceConfig
from ptp.ordinal import Ordinal
import re
from datetime import datetime

class Dblp(EventDataSource):
    '''
    scientific events from https://dblp.org
    '''
    sourceConfig = EventDataSourceConfig(lookupId="dblp", name="dblp", url='https://dblp.org/', title='dblp computer science bibliography', tableSuffix="dblp",locationAttribute="location")
    
    def __init__(self):
        '''
        constructor
        '''
        super().__init__(DblpEventManager(), DblpEventSeriesManager(), Dblp.sourceConfig)
       
    @classmethod
    def setPatterns(cls):
        if hasattr(cls,"dayPattern"):
            return
        cls.dayPattern=r""
        cls.dayPatternWithSuffix=r""
        delim=""
        for day in range(1,31):
            # https://stackoverflow.com/a/739266/1497139
            cls.dayPattern=cls.dayPattern+f"{delim}{day}"
            delim="|"
        cls.monthPattern=r"January|February|March|April|May|June|July|August|September|October|November|December"
        cls.yearPattern=r"([12][0-9]{3})"
        cls.ws=r"\s+"
        cls.dateRangePattern=f"^\s*({cls.dayPattern})[-]({cls.dayPattern})(st|nd|rd|th)?{cls.ws}({cls.monthPattern}){cls.ws}({cls.yearPattern})\s*[.]?$"        


    @classmethod
    def strToDate(cls,dateStr):
        '''
        Args:
            dateStr(str): the string to convert
            
        Return:
            datetime: the date
        '''
        d=None
        try:
            d = datetime.strptime(dateStr, '%d %B %Y')
        except ValueError as _ve:
            
            pass
        return d
    
    @classmethod
    def getDateRangeFromTitle(cls,title:str):
        dateRange={}
        if title is not None:
            parts=title.split(",")
            datePartStr=parts[len(parts)-1]
            dateRange=Dblp.getDateRange(datePartStr)
        return dateRange 
    
    @classmethod
    def getDateRange(cls,dateStr):
        '''
        given a dblp date string create a date range
        
        Args:
            dateStr(str): the date string to analyze
            
        Returns:
            dict: containing year, startDate, endDate
        examples:
            18-21 September 2005

        '''
        cls.setPatterns()
        result={}
        if dateStr is not None:
            yearOnly=re.search(f"^{cls.yearPattern}$",dateStr)
            dateRangeMatch=re.search(cls.dateRangePattern,dateStr)
            if yearOnly: 
                result['year']=int(yearOnly.group(1))
            elif dateRangeMatch:      
                fromDay=dateRangeMatch.group(1)
                toDay=dateRangeMatch.group(2)
                month=dateRangeMatch.group(4)
                year=dateRangeMatch.group(5)
                startDateStr=f"{fromDay} {month} {year}"
                toDateStr=f"{toDay} {month} {year}"   
                startDate=cls.strToDate(startDateStr)   
                endDate=cls.strToDate(toDateStr)
                if startDate and endDate: 
                    result['startDate']=startDate
                    result['endDate']=endDate
        if 'startDate' in result:
                result['year']=result['startDate'].year
        return result

        
class DblpEvent(Event):
    '''
    a Dblp Event
    
    Example event: https://dblp.org/db/conf/aaai/aaai2020.html
    '''

    def __init__(self):
        '''constructor '''
        super().__init__()
        pass
        
    
    @staticmethod
    def postProcessLodRecord(rawEvent:dict):
        '''
        fix the given raw Event
        
        Args:
            rawEvent(dict): the raw event record to fix
        '''
        if 'url' in rawEvent:
            rawEvent["url"] = f"https://dblp.org/{rawEvent['url']}" 
        if 'title' in rawEvent:   
            title=rawEvent["title"]
            dateRange=Dblp.getDateRangeFromTitle(title)
            for key, value in dateRange.items():
                rawEvent[key]=value
            Ordinal.addParsedOrdinal(rawEvent)
                
        if "year" in rawEvent:
            # set year to integer value
            yearStr = rawEvent['year']
            year = None
            try:
                year = int(yearStr)
            except Exception as _ne:
                pass
            rawEvent['year'] = year
            # if there is a booktitle create acronym
            if "booktitle" in rawEvent:
                booktitle = rawEvent['booktitle']
                if booktitle is not None and year is not None:
                    acronym = f"{booktitle} {year}"
                    rawEvent["acronym"] = acronym 
        doiprefix = "https://doi.org/"
        if 'ee' in rawEvent:
            ees = rawEvent['ee']
            if ees:
                for ee in ees.split(","):
                    if ee.startswith(doiprefix):
                        doi = ee.replace(doiprefix, "")
                        rawEvent["doi"] = doi 


class DblpEventSeries(EventSeries):
    '''
    a Dblp Event Series
    
    Example event series: https://dblp.org/db/conf/aaai/index.html
    '''

    def __init__(self):
        '''constructor '''
        super().__init__()
        pass
        

class DblpEventManager(EventManager):
    '''
    dblp event access (in fact proceedings)
    
    Example event: https://dblp.org/db/conf/aaai/aaai2020.html
    
    '''
    cacheOnly=False

    def __init__(self, config: StorageConfig=None):
        '''
        Constructor
        '''
        super(DblpEventManager, self).__init__(name="DblpEvents", sourceConfig=Dblp.sourceConfig, clazz=DblpEvent, config=config)

    pass

    def configure(self):
        '''
        configure me
        '''
        withProgress = False
        if DblpEventManager.cacheOnly:
            return
        if not hasattr(self, "dblpXml"): 
            self.dblpXml = DblpXml()
            self.dblpXml.warnFullSize()
            withProgress = True
        self.sqlDb = self.dblpXml.getXmlSqlDB(showProgress=withProgress)
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts = self.getLoDfromDblp

    def getLoDfromDblp(self) -> list:
        '''
        get the LoD for the event series
            
        Return:
            list: the list of dict with my series data

        '''
        query = """select conf as series,title,year,url,booktitle,series as publicationSeries,ee,isbn,mdate,key as eventId
        from proceedings 
        order by series,year"""
        listOfDicts = self.sqlDb.query(query)
        self.setAllAttr(listOfDicts, "source", "dblp")
        self.postProcessLodRecords(listOfDicts)
        return listOfDicts
    
    def addLocations(self):
        # extract locations
        for dblpEvent in self.events:
            title=dblpEvent.title
            if title is not None:
                parts=title.split(",")
                if len(parts)>3:
                    dblpEvent.location=f"{parts[2].strip()}, {parts[3].strip()}"

class DblpEventSeriesManager(EventSeriesManager):
    '''
    dblp event series access
    Example event series: https://dblp.org/db/conf/aaai/index.html

    dblp provides regular dblp xml dumps
    '''

    def __init__(self, config: StorageConfig=None):
        '''
        Constructor
        '''
        super().__init__(name="DblpEventSeries", sourceConfig=Dblp.sourceConfig, clazz=DblpEventSeries, config=config)
        
    def configure(self):
        '''
        configure me
        '''
        withProgress = False
        if DblpEventManager.cacheOnly:
            return
        if not hasattr(self, "dblpXml"): 
            self.dblpXml = DblpXml()
            self.dblpXml.warnFullSize()
            withProgress = True
        self.sqlDb = self.dblpXml.getXmlSqlDB(showProgress=withProgress)
        if not hasattr(self, "getListOfDicts"):
            self.getListOfDicts = self.getLoDfromDblp

    def getLoDfromDblp(self) -> list:
        '''

        get the list of dicts for the event data
            
        Return:
            list: the list of dict with my event data

        '''
        query = """select conf as acronym,conf as eventSeriesId,count(*) as count,min(year) as minYear,max(year) as maxYear
        from proceedings 
        where acronym is not null
        group by acronym
        order by 2 desc"""
        listOfDicts = self.sqlDb.query(query)
        self.setAllAttr(listOfDicts, "source", "dblp")
        self.postProcessLodRecords(listOfDicts)
        return listOfDicts




import pathlib
import os
from setuptools import setup
from collections import OrderedDict
from corpus.version import Version

try:
    long_description = ""
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

except:
    print('Curr dir:', os.getcwd())
    long_description = open('../../README.md').read()
here = pathlib.Path(__file__).parent.resolve()
requirements = (here / 'requirements.txt').read_text(encoding='utf-8').split("\n")

setup(name='ConferenceCorpus',
      version=Version.version,
      description='python api providing access to academic events and event series from different sources',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://wiki.bitplan.com/index.php/ConferenceCorpus',
      download_url='https://github.com/WolfgangFahl/ConferenceCorpus',
      author='Wolfgang Fahl',
      author_email='wf@bitplan.com',
      license='Apache',
      project_urls=OrderedDict(
        (
            ("Code", "https://github.com/WolfgangFahl/ConferenceCorpus"),
            ("Issue tracker", "https://github.com/WolfgangFahl/ConferenceCorpus/issues"),
        )
      ),
      classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10'
      ],
      packages=['resources','corpus','corpus.datasources','corpus.utils','corpus.quality','corpus.xmlhandler', 'corpus.matching','ptp'],
      package_data={'resources':['*.yaml', '*.json', 'matching/*.json']},
      install_requires=requirements,
      entry_points={
         'console_scripts': [
             'aelookup = corpus.lookup:main',
             'ccTibkat = corpus.datasources.tibkatftx:main',
             'ccUpdate = corpus.ccupdate:main'
      ],
    },
      zip_safe=False)



"""
created at 2022-08-08
updated at 2022-08-08
@author: th
"""

import itertools
import json
import pathlib
from enum import Enum

import matplotlib.pyplot as plt
from typing import List, Tuple, Dict
from sklearn import metrics
from sklearn.metrics import precision_score, recall_score, accuracy_score
from dataclasses import dataclass

class MatchEvaluation:
    """
    provides method to analyze event matchings
    """
    source_ids = ['openresearchId', 'wikicfpId', 'wikidataId', 'TibKatId', 'dblpConferenceId', 'dblpSeriesId',
                  'gndId', 'seriesWikidataId', 'seriesWikicfpId', 'seriesOpenresearchId', 'confrefId',
                  'confrefSeriesId']

    def __init__(self, training_data: Dict[str, Dict[str, List[str]]]):
        """
        constructor

        Args:
            training_data: for every source an id lookup for the matches to other sources
        """
        self.lut = training_data
        self.evaluation_result: List[Tuple[bool, bool]] = []
        self.results: List[MatchingResult] = []

    @property
    def actual(self):
        """
        Ground truth (correct) target values
        """
        return [actual for actual, _ in self.evaluation_result]

    @property
    def predicted(self):
        """
        Estimated targets (True if match is valid)
        """
        return [predicted for _, predicted in self.evaluation_result]

    @property
    def precision(self) -> float:
        """
        Compute the precision.
        """
        return precision_score(self.actual, self.predicted)

    @property
    def recall(self) -> float:
        """
        Compute the recall.
        """
        return recall_score(self.actual, self.predicted)

    @property
    def accuracy(self) -> float:
        """
        Accuracy classification score.
        """
        return accuracy_score(self.actual, self.predicted)

    def print_scores(self):
        """
        prints the evaluation scores
        """

        print('Precision: %.3f' % self.precision)
        print('Recall: %.3f' % self.recall)
        print('Accuracy: %.3f' % self.accuracy)

    def get_results_classified_as(self, classifications: List['CM']) -> List['MatchingResult']:
        """

        Args:
            classifications: result classifications that should be printed out

        Returns:
            list of matching results
        """
        res = [mr for mr in self.results if mr.classification in classifications]
        return res

    def plot_matching_result(self, classifications: List['CM'] = None):
        """
        generates a graphviz plot of the matching result
        Args:
            classifications: result classifications that should be included in the plot

        Returns:

        """
        results = self.results
        if classifications is not None:
            results = self.get_results_classified_as(classifications)
        plot_body = ""
        id_lut = {}
        for result in results:
            id_lut[result.id_1] = (result.id_1.__hash__(), result.source_1)
            id_lut[result.id_2] = (result.id_2.__hash__(), result.source_2)
        # add all nodes
        for label, (node_name, source) in id_lut.items():
            plot_body += f"""{node_name} [label="{label}"]\n"""  # ToDo: add color
        for result in results:
            color = ""
            if result.classification == CM.FalseNegative:
                color = "orange"
            elif result.classification == CM.TrueNegative:
                color = "red"
            else:
                color = "green"
            relation_att = f"""[ color="{color}"]"""
            plot_body += f"""{id_lut[result.id_1][0]} -- {id_lut[result.id_2][0]} {relation_att};\n"""
        legend = f"""subgraph cluster_01 {{
    label = "Legend";
    node [shape=point]
    {{
        rank=same
        d0 [style = invis];
        d1 [style = invis];
        p0 [style = invis];
        p1 [style = invis];
        s0 [style = invis];
        s1 [style = invis];
    }}
    d0 -- d1 [label=correct color=green]
    p0 -- p1 [label="incorrect/\nnot in source of truth" color=red]
    s0 -- s1 [label=missing color=orange]
}}"""
        plot = f"""graph untitl {{\n rankdir=LR;\n {plot_body}\n{legend}\n}}"""
        with open("/tmp/matching_result.txt", mode="w") as fp:
            print(plot, file=fp)


    @staticmethod
    def convert_to_labeled_matching_result(result: List[bool]):
        """
        converts the given matching results to a labeled list
        """
        return ["match" if match else "no-match" for match in result]

    def plot_confusion_matrix(self, show: bool=False):
        """
        prints out different evaluation metrics
        """
        y_true = self.convert_to_labeled_matching_result(self.actual)
        pred = self.convert_to_labeled_matching_result(self.predicted)
        metrics.ConfusionMatrixDisplay.from_predictions(y_true, pred, cmap=plt.cm.Blues)
        plt.xlabel('Prediction', fontsize=18)
        plt.ylabel('Actual', fontsize=18)
        plt.title('Confusion Matrix', fontsize=18)
        if show:
            plt.show()

    def evaluate_records(self, records: List[dict], source_id_names: List[str] = None):
        """
        evaluates the given records and checks id the matchings are correct
        Args:
            records: list of event records
            source_id_names: names of the source ids which should be evaluated. the names must be included in the source_ids
        """
        if source_id_names is None:
            source_id_names = self.source_ids
        self.normalize_records(records)
        id_lut = {k: set(v.keys()) for k, v in self.lut.items()}
        for record in records:
            for i, source_id_1 in enumerate(source_id_names):
                for source_id_2 in source_id_names[i:]:
                    v1_ids = self.get_ids_from_record(record, source_id_1)
                    v2_ids = self.get_ids_from_record(record, source_id_2)
                    if source_id_1 == source_id_2:
                        continue
                    # check present matches
                    for v1, v2 in itertools.product(v1_ids, v2_ids):
                        if v1 is not None and v2 is not None:
                            expected = True
                            actual = self.check_matching(source_id_1, v1, source_id_2, v2)
                            if actual:
                                self.results.append(MatchingResult(CM.TruePositive, source_id_1, v1, source_id_2, v2))
                            else:
                                self.results.append(MatchingResult(CM.TrueNegative, source_id_1, v1, source_id_2, v2))
                            self.evaluation_result.append((actual, expected))
                    if "series" in source_id_1.lower():
                        # series is only checked from the other direction
                        continue
                    # check for missing matches
                    for no_match_id in id_lut[source_id_2].difference(v2_ids):
                        expected = False
                        for v1 in v1_ids:
                            actual = self.check_matching(source_id_1, v1, source_id_2, no_match_id)
                            if actual:
                                self.results.append(
                                        MatchingResult(CM.FalseNegative, source_id_1, v1, source_id_2, no_match_id))
                            else:
                                self.results.append(
                                    MatchingResult(CM.FalsePositive, source_id_1, v1, source_id_2, no_match_id))
                            self.evaluation_result.append((actual, expected))

    @staticmethod
    def get_ids_from_record(record: dict, source_id_name:str, separator: str = "⇹") -> List[str]:
        """
        extract and convert the value for the given source id property

        Args:
            record: event record
            source_id_name: name of the source id property
            separator: id separator

        Returns:
            list of ids for the given source_id_name
        """
        if separator is None:
            separator = "⇹"
        res = record.get(source_id_name, [])
        if res is None:
            res = []
        if isinstance(res, str):
            res = res.split(separator)
        return res


    def evaluate_present_matches(self, ):
        """
        evaluates explicitly stated matchings
        """
        pass

    def check_matching(self, source1: str, id1: str, source2: str, id2: str):
        """
        Check if the given matching is correct
        Args:
            source1: source id name
            id1: id value of the first source
            source2: source id name
            id2: id value of the second source

        Returns:
            True if the given matching is correct. Otherwise, False is returned.
        """
        try:
            valid_match = id2 in self.lut.get(source1, {}).get(id1, {}).get(source2,  [])
            valid_match = valid_match or id1 in self.lut.get(source2, {}).get(id2, {}).get(source1, [])
        except Exception:
            valid_match = False
        return valid_match

    @staticmethod
    def normalize_records(records: List[dict]):
        """
        normalizes the source id property names
        Args:
            records: list of event records
        """
        for record in records:
            source = record.get("source")
            event_id = record.get("eventId")
            series_id = record.get("seriesId")
            if source == "confref":
                record['confrefId'] = event_id
                record['confrefSeriesId'] = series_id
            elif source == "dblp":
                record["dblpConferenceId"] = event_id
                if series_id is not None:
                    record["dblpSeriesId"] = "conf/" + series_id
            elif source == "tibkat":
                record["TibKatId"] = event_id
                record["gndId"] = record.get("gndIds")
            elif source == "wikicfp":
                record["wikicfpId"] = event_id
                record["seriesWikicfpId"] = series_id
            elif source == "wikidata":
                record["wikidataId"] = event_id
                record["seriesWikidataId"] = record.get("eventInSeriesId")
                record["dblpConferenceId"] = record.get("dblpId")
                record["wikicfpId"] = record.get("wikiCfpId")
            elif source in ["orclone", "openresearch"]:
                record["openresearchId"] = event_id
                record["seriesOpenresearchId"] = record.get("inEventSeries")
                dblp_conference_id = record.get("dblpId")
                if dblp_conference_id is not None and not dblp_conference_id.startswith("conf/"):
                    dblp_conference_id = "conf/" + dblp_conference_id
                record["dblpConferenceId"] = dblp_conference_id

    def group_ids_by_source(self, records: List[dict], source_id_names: List[str] = None) -> Dict[str, set]:
        """
        for each source id in self.sourceIds provide a set of all used Ids
        Args:
            records: list of event records

        Returns:
            Returns for each sourceId the set of all used ids
        """
        if source_id_names is None:
            source_id_names = self.source_ids
        res = {}
        for sourceId in source_id_names:
            ids = set()
            for record in records:
                if record.get(sourceId) is not None:
                    value = record.get(sourceId, [])
                    if isinstance(value, list):
                        ids.union(value)
                    else:
                        ids.add(value)
            res[sourceId] = ids
        return res

    @classmethod
    def fromResources(cls) -> 'MatchEvaluation':
        """
        Load MatchEvaluation with training data from the resources
        """
        grandParent = pathlib.Path(__file__).parent.parent.parent.resolve()
        filepath = (grandParent / "resources/matching/training_set.json")
        path = str(filepath)
        with open(path, mode="r") as fp:
            lod = json.load(fp)
        lut = {}
        # construct lookup table for source ids
        for source_id in MatchEvaluation.source_ids:
            lut[source_id] = {}
            for record in lod.values():
                values = record.get(source_id)
                if values is None or values == []:
                    continue
                for value in values:
                    lut[source_id][value] = record
        matchEvaluation = MatchEvaluation(training_data=lut)
        return matchEvaluation


@dataclass
class MatchingResult:
    """
    represents a matching result
    """
    classification: 'CM'
    source_1: str
    id_1: str
    source_2: str
    id_2: str

class CM(Enum):
    """
    results of the confusion matrix
    """
    TruePositive = "TP"
    FalsePositive = "FP"
    TrueNegative = "TN"
    FalseNegative = "FN"